# -*- coding: utf-8 -*-
import os
import sys
import base64
import time

from comp_exceptions import (
    OrderWaitingProcessingException,
    ErrorProcessing,
)
from fiscalpersistence import Tender
from typing import Optional, Dict, List

import msgbus
import six
import logging
import sysactions

from sysactions import set_custom
from fiscalinterface import FiscalProcessor
from xml.etree import cElementTree as eTree
from threading import Thread, Condition
from msgbus import MBException, FM_STRING, FM_PARAM, TK_SYS_ACK, MBEasyContext
from syserrors import SE_NOTFOUND, SE_MB_CONNECT
from systools import sys_log_exception
from bustoken import (
    TK_SAT_PROCESS_REQUEST,
    TK_SAT_STATUS_REQUEST,
    TK_SAT_CANCEL_SALE,
)
from old_helper import convert_from_localtime_to_utc
from pos_model import OrderParser
from satbuilder import CfeBuilder
from datetime import datetime
from common import FiscalParameterController
from pos_util import SaleLineUtil

logger = logging.getLogger("FiscalWrapper")


class SatRequestBuilder:

    def __init__(
        self,                           # type: SatRequestBuilder
        mbcontext,                      # type: MBEasyContext
        crt,                            # type: int
        cnpj_sw,                        # type: str
        sign_ac,                        # type: str
        cnpj_contribuinte,              # type: str
        inscr_estadual,                 # type: str
        order_parser,                   # type: OrderParser
        fiscal_parameter_controller,    # type: FiscalParameterController
        sale_line_util,                 # type: SaleLineUtil
        cfe_builder,                    # type: CfeBuilder
        is_mfe,                         # type: bool
    ):
        # type: (...) -> None

        self.crt = crt
        self.mbcontext = mbcontext
        self.cnpj_sw = cnpj_sw
        self.sign_ac = sign_ac
        self.cnpj_contribuinte = cnpj_contribuinte
        self.inscr_estadual = inscr_estadual
        self.order_parser = order_parser
        self.fiscal_parameter_controller = fiscal_parameter_controller
        self.sale_line_util = sale_line_util
        self.cfe_builder = cfe_builder
        self.is_mfe = is_mfe

    def get_sat_info(self):
        return self.cnpj_sw, self.sign_ac

    def build_request(self, order_xml):
        # type: (eTree) -> unicode
        order = self.order_parser.parse_order(order_xml)

        context = {}
        xml = self.cfe_builder.build_xml(order, context)
        return xml


class SatService:
    def __init__(self, service_name):
        self.service_name = service_name
        self.available = True


class SatServiceFinder:
    def __init__(self, mbcontext, max_sat_number, pos_sat_list, sleep_time):
        # type: (MBEasyContext, int, str, int) -> None
        self.mbcontext = mbcontext

        self.max_sat_number = max_sat_number
        self.pos_sat_list = pos_sat_list.split(';') if pos_sat_list else []
        self.sleep_time = sleep_time

        self.search_thread = None  # type: Optional[Thread]
        self.search_thread_condition = Condition()
        self.search_thread_running = False
        self.available_sats = {}  # type: Dict[str, SatService]
        self.available_sats_keys = []
        self.next_sat_index = 0
        self.available_sats_lock = Condition()
        self.processing_orders_id = []  # type: List[str]

        self.service_name_initial = "SAT"

        self.search_sat_modules()
        self.start_search_thread()

    def find_and_lock_sat_service(
        self,      # type: SatServiceFinder
        order_id,  # type: str
    ):
        # type: (...) -> str

        logger.debug("Find and_lock_sat_service for OrderId: {}".format(order_id))

        with self.available_sats_lock:
            if not self.available_sats:
                with self.search_thread_condition:
                    self.search_thread_condition.notify()
                raise Exception("Nenhum SAT encontrado")

            initial_sat_index = self.next_sat_index

            try:
                sat_service_name = self.available_sats_keys[self.next_sat_index]
            except IndexError:
                log_message = "Erro ao acessar SAT com indice {}. Sats disponiveis: {}"
                logger.exception(log_message.format(self.next_sat_index, len(self.available_sats_keys)))
                self.next_sat_index = 0
                initial_sat_index = len(self.available_sats_keys)
                sat_service_name = self.available_sats_keys[self.next_sat_index]
            else:
                self.next_sat_index += 1
                if self.next_sat_index >= len(self.available_sats):
                    self.next_sat_index = 0

            time_limit = (time.time() + 95)
            while not self.available_sats[sat_service_name].available:
                sat_service_name = self.available_sats_keys[self.next_sat_index]
                self.next_sat_index += 1
                if self.next_sat_index >= len(self.available_sats):
                    self.next_sat_index = 0

                if self.next_sat_index == initial_sat_index:
                    self.available_sats_lock.wait(timeout=5)
                    if time.time() > time_limit:
                        logger.warning("Reached time limit while finding an available SAT")
                        raise Exception("Nenhum SAT encontrado")

                    initial_sat_index = self.next_sat_index
                    sat_service_name = self.available_sats_keys[self.next_sat_index]

            self.available_sats[sat_service_name].available = False
            return sat_service_name

    def unlock_sat_service(self, sat_service_name):
        # type: (str) -> None
        with self.available_sats_lock:
            if sat_service_name in self.available_sats:
                self.available_sats[sat_service_name].available = True

                # Como liberamos um SAT, vamos notificar para caso alguem esteja esperando
                self.available_sats_lock.notify()

    def sat_unavailable(self, sat_service_name):
        # type: (str) -> None
        if len(self.available_sats) == 1:
            return
        with self.available_sats_lock:
            if sat_service_name in self.available_sats:
                del self.available_sats[sat_service_name]
                self.available_sats_keys.remove(sat_service_name)
                posid = int(sat_service_name[3:])
                set_custom(posid, 'SAT_STATUS', '2')

                # Removemos um SAT - Vamos arrumar os indices
                self._fix_sat_index()

    def set_order_processing(
        self,       # type: SatServiceFinder
        order_id,   # type: str
    ):
        # type: (...) -> None

        with self.available_sats_lock:
            self.processing_orders_id.append(order_id)

    def is_order_waiting_processing(
        self,       # type: SatServiceFinder
        order_id,   # type: str
    ):
        # type: (...) -> bool

        with self.available_sats_lock:
            return order_id in self.processing_orders_id

    def remove_order_processing(
        self,       # type: SatServiceFinder
        order_id,   # type: str
    ):
        # type: (...) -> None

        with self.available_sats_lock:
            if order_id in self.processing_orders_id:
                logger.debug("Remove order processing - OrderId: {}".format(order_id))
                self.processing_orders_id.remove(order_id)

    def search_sat_modules(self):
        logger.debug("Start Search SAT Function")
        sat_list_to_search = self.pos_sat_list if self.pos_sat_list else range(0, self.max_sat_number + 1)

        now_available = {}
        for i in sat_list_to_search:
            sat_service_name = "SAT%s" % str(i).zfill(2)

            try:
                ret = self.mbcontext.MB_EasySendMessage(
                    dest_name=sat_service_name,
                    token=TK_SAT_STATUS_REQUEST,
                    format=FM_PARAM,
                    timeout=100 * 100000,
                )
                if ret.data == "OK":
                    now_available[sat_service_name] = SatService(sat_service_name)
                    try:
                        set_custom(i, 'SAT_STATUS', '1') if i != 0 else None
                    except Exception:
                        pass
                else:
                    if ret.data != "No SAT attached":
                        logger.warning("{0} not available: {1}".format(sat_service_name, ret.data))
                    try:
                        set_custom(i, 'SAT_STATUS', '2') if i != 0 else None
                    except Exception:
                        pass
            except MBException as e:
                try:
                    set_custom(i, 'SAT_STATUS', '3') if i != 0 else None
                except Exception:
                    pass
                if e.errorcode not in [SE_NOTFOUND, SE_MB_CONNECT]:
                    logger.exception(e.message)

        with self.available_sats_lock:
            # Primeiro removemos os que estavam disponiveis e agora nao estao mais
            sats_to_remove = []
            for sat_service_name in self.available_sats:  # type: str
                if sat_service_name not in now_available:
                    sats_to_remove.append(sat_service_name)

            for sat_service_name in sats_to_remove:
                del self.available_sats[sat_service_name]
                self.available_sats_keys.remove(sat_service_name)

            # Agora adicionamos os que passaram a estar disponiveis
            sats_novos = False
            for sat_service_name in now_available:  # type: str
                if sat_service_name not in self.available_sats:
                    sat = now_available[sat_service_name]  # type: SatService

                    self.available_sats[sat.service_name] = sat
                    self.available_sats_keys.append(sat.service_name)

                    sats_novos = True

            # Removemos um SAT - Vamos arrumar os indices
            self._fix_sat_index()

            if sats_novos:
                # Como temos SATs novos disponiveis, vamos avisar caso algume thread esteja esperando
                self.available_sats_lock.notify()
        logger.debug("End Search SAT Function")

    def start_search_thread(self):
        self.search_thread_running = True
        self.search_thread = Thread(target=self._search_thread_loop)
        self.search_thread.daemon = True
        self.search_thread.start()

    def stop_search_thread(self):
        if self.search_thread_running:
            with self.search_thread_condition:
                self.search_thread_running = False
                self.search_thread_condition.notifyAll()

    def _search_thread_loop(self):
        while self.search_thread_running:
            with self.search_thread_condition:
                if not self.available_sats:
                    self.search_thread_condition.wait(15)
                else:
                    self.search_thread_condition.wait(self.sleep_time)
            self.search_sat_modules()

    def _fix_sat_index(self):
        # Atualizamos o indice, pois houve alguma remoção de SATs
        if self.next_sat_index >= len(self.available_sats):
            self.next_sat_index = len(self.available_sats) - 1

            # Não permirte que o índice fique negativo
            if self.next_sat_index < 0:
                self.next_sat_index = 0


class MfeServiceFinder(SatServiceFinder):
    def find_and_lock_sat_service(self, order_id):
        # type: (str) -> str
        with self.available_sats_lock:
            if not self.available_sats:
                raise Exception("Nenhum SAT encontrado")

            sat_service_name = self.available_sats_keys[int(order_id) % len(self.available_sats_keys)]

            while not self.available_sats[sat_service_name].available:
                self.available_sats_lock.wait()

            self.available_sats[sat_service_name].available = False
            return sat_service_name


class SatProcessor(FiscalProcessor):
    def __init__(self, mbcontext, sat_service_finder, sat_request_builder, fiscal_sent_dir):
        # type: (msgbus.MBEasyContext, SatServiceFinder, SatRequestBuilder, str) -> SatProcessor
        super(SatProcessor, self).__init__()

        self.mbcontext = mbcontext
        self.sat_service_finder = sat_service_finder
        self.sat_request_builder = sat_request_builder
        self.last_sat_used = {}
        self.fiscal_sent_dir = fiscal_sent_dir
        sysactions.mbcontext = self.mbcontext
        self.cancel_data = None

    def get_sat_info(self):
        return self.sat_request_builder.get_sat_info()

    def request_fiscal(
            self,               # type: SatProcessor
            pos_id,             # type: str
            order,              # type: eTree
            tenders,            # type: List[Tender]
            paf=False,          # type: Optional[bool]
            fiscal_id=None      # type: str
    ):
        # type: (...) -> eTree.Element

        order_id = order.get("orderId")
        order_waiting_processing = self.sat_service_finder.is_order_waiting_processing(
            order_id=order_id,
        )
        if order_waiting_processing:
            log_message = "Order waiting processing PosId: {} - OrderId: {}"
            logger.debug(log_message.format(pos_id, order_id))
            raise OrderWaitingProcessingException()

        sat_service_name = None
        try:
            log_message = "Set order processing PosId: {} - OrderId: {}"
            logger.debug(log_message.format(pos_id, order_id))

            self.sat_service_finder.set_order_processing(
                order_id=order_id,
            )

            log_message = "Generating SAT xml request PosId: {} - OrderId: {}"
            logger.debug(log_message.format(pos_id, order_id))

            request = self.sat_request_builder.build_request(
                order_xml=order,
            )

            sat_service_name = self.sat_service_finder.find_and_lock_sat_service(
                order_id=order_id,
            )

            self.last_sat_used[pos_id] = sat_service_name

            dir_sent = self._get_sent_directory()

            log_message = "Sent SAT xml processing PosId: {} - OrderId: {}"
            logger.debug(log_message.format(pos_id, order.get("orderId")))

            request_data = "\0".join(
                [
                    pos_id,
                    order.attrib["orderId"],
                    request.encode("utf-8"),
                    fiscal_id,
                ]
            )
            msg = self.mbcontext.MB_EasySendMessage(
                dest_name=sat_service_name,
                token=TK_SAT_PROCESS_REQUEST,
                format=FM_STRING,
                data=request_data,
            )

            if msg.token == TK_SYS_ACK:
                ret_xml = eTree.XML(msg.data)
                self._process_return_fiscalization(
                    pos_id=pos_id,
                    order_id=order_id,
                    dir_sent=dir_sent,
                    ret_xml=ret_xml,
                    sat_service_name=sat_service_name,
                )
                return ret_xml

            logger.debug("Fail processing SAT xml PosId: {} - OrderId: {}".format(pos_id, order_id))
            self._save_invoice_error_processing(
                pos_id=pos_id,
                order_id=order_id,
                request=request,
                sat_service_name=sat_service_name,
            )
            raise ErrorProcessing(msg.data)

        except Exception as e:
            sys_log_exception(e.message)
            if sat_service_name:
                self.sat_service_finder.sat_unavailable(
                    sat_service_name=sat_service_name,
                )
            raise six.reraise(Exception, e, sys.exc_info()[2])
        finally:
            self.sat_service_finder.remove_order_processing(
                order_id=order_id,
            )

    def check_nf_situation(
        self,           # type: SatProcessor
        initial_date,   # type: str
        final_date,     # type: str
    ):
        # type: (...) -> (bool, str)

        return False, "Funcionalidade exclusiva para o modo NFCe"

    def do_validation(self, get_days_to_expiration=None):
        return True, "OK"

    def get_fiscal_sent_dir(self):
        return self.fiscal_sent_dir

    def terminate(self):
        self.sat_service_finder.stop_search_thread()

    def cancel_sale(
        self,
        order_data,     # type: str
        order_id,       # type: str
    ):
        # type: (...) -> Dict

        self.cancel_data = None
        sat_service = "SAT" + order_data[0]
        cancel_data = self.build_cancel_xml(order_data[1])
        msg = self.mbcontext.MB_EasySendMessage(
            dest_name=sat_service,
            token=TK_SAT_CANCEL_SALE,
            format=FM_PARAM,
            timeout=100 * 100000,
            data=str(cancel_data),
        )
        if msg.token == TK_SYS_ACK:
            self.cancel_data = msg.data
            ret_xml = eTree.XML(msg.data)
            element = ret_xml.find("XmlResponse").text
            message = ret_xml.find("Message").text
            if element:
                padded = element + "=" * ((4 - len(element) % 4) % 4)
                base64_xml = base64.b64decode(padded)
                cfe = cancel_data[1]
                num_nota_cfe = cfe[34:40].zfill(9)
                idx_start = base64_xml.find("<dEmi>") + 6
                idx_final = base64_xml.find("</dEmi>")
                data_emissao = base64_xml[idx_start:idx_final]
                dir_arquivo = os.path.join(data_emissao[0:4], data_emissao[4:6], data_emissao[6:8])
                pos_id = sat_service[-2:]
                file_name = "S{}_{}_{}_cancelamento_{}.xml".format(pos_id, num_nota_cfe, order_id.zfill(9), cfe)
                file_path = os.path.join(self.fiscal_sent_dir, "Enviados", dir_arquivo, file_name)
                with open(file_path, "w+") as cfe_proc_file:
                    cfe_proc_file.write(base64_xml)

                return {'status': True, 'message': message, 'xml_response': padded}

            return {'status': True, 'message': message, 'xml_response': None}
        else:
            return {'status': False, 'message': msg.data, 'xml_response': None}

    def build_cancel_xml(self, order_request):
        sale_request_xml = eTree.XML(base64.b64decode(order_request + "==="))

        chcanc = sale_request_xml.find("infCFe").attrib.get("Id")
        cnpj = self.sat_request_builder.cnpj_sw
        signac = self.sat_request_builder.sign_ac
        posid = sale_request_xml.find("infCFe").find("ide").find("numeroCaixa").text

        cancel_xml = "<CFeCanc>" \
                     "<infCFe chCanc=\"{chcanc}\">" \
                     "<ide>" \
                     "<CNPJ>{cnpj}</CNPJ>" \
                     "<signAC>{signac}</signAC>" \
                     "<numeroCaixa>{posid}</numeroCaixa>" \
                     "</ide>" \
                     "<emit/>" \
                     "<dest/>" \
                     "<total/>" \
                     "</infCFe>" \
                     "</CFeCanc>" \
            .format(chcanc=chcanc, cnpj=cnpj, signac=signac, posid=posid)
        return cancel_xml, chcanc

    def _process_return_fiscalization(
        self,               # type: SatProcessor
        pos_id,             # type: str
        order_id,           # type: str
        dir_sent,           # type: str
        ret_xml,            # type: eTree.Element
        sat_service_name,   # type: str
    ):
        # type: (...) -> None

        element = ret_xml.find("XmlRequest").text
        padded = element + "=" * ((4 - len(element) % 4) % 4)

        logger.debug("PosId: {} - OrderId: {} - request size: {}".format(pos_id, order_id, len(padded)))
        base64_xml = base64.b64decode(padded)
        logger.debug("PosId: {} - OrderId: {} - base64_xml: {}".format(pos_id, order_id, base64_xml))

        req_xml = eTree.XML(base64_xml)
        date_emission = req_xml.find("infCFe/ide/dEmi").text
        time_emission = req_xml.find("infCFe/ide/hEmi").text

        dir_nota = self._get_invoice_directory(
            date_emission=date_emission,
            dir_sent=dir_sent,
        )

        num_nota_cfe_el = req_xml.find("infCFe/ide/nCFe")
        num_nota_cfe = num_nota_cfe_el.text.zfill(9) if num_nota_cfe_el is not None else "000000000"

        log_message = "Success processing SAT Xml - PosId: {} - OrderId: {} - FiscalId: {}"
        logger.debug(log_message.format(pos_id, order_id, num_nota_cfe))

        self._set_order_fiscalization_custom_properties(
            pos_id=pos_id,
            order_id=order_id,
            date_string=date_emission + time_emission,
            num_nota_cfe=num_nota_cfe,
        )

        self.sat_service_finder.unlock_sat_service(
            sat_service_name=sat_service_name,
        )

        cfe_code = ret_xml.find("FiscalData").find("satkey").text
        file_name = "S{0}_{1}_{2}_response_pos{3}_CFe{4}.xml".format(
            sat_service_name[-2:],
            num_nota_cfe,
            order_id.zfill(9),
            str(pos_id).zfill(2),
            cfe_code
        )

        response_file_dir = os.path.join(
            dir_nota,
            file_name,
        )

        log_message = "Salve SAT xml PosId: {} - OrderId: {} - response_file_dir: {}"
        logger.debug(log_message.format(pos_id, order_id, response_file_dir))
        with open(response_file_dir, "w+") as response_file:
            response_file.write(base64_xml)

    def _set_order_fiscalization_custom_properties(
        self,           # type: SatProcessor
        pos_id,         # type: str
        order_id,       # type: str
        date_string,    # type: str
        num_nota_cfe,   # type: str
    ):
        # type: (...) -> None

        local_time = datetime.strptime(
            date_string,
            "%Y%m%d%H%M%S",
        )
        emission_date = convert_from_localtime_to_utc(
            localtime=local_time,
        )
        custom_properties = {
            "FISCALIZATION_DATE": emission_date.strftime("%Y-%m-%dT%H:%M:%S"),
            "FISCAL_ID": str(num_nota_cfe),
        }
        self.order_service.set_custom_properties(
            pos_id=pos_id,
            custom_properties=custom_properties,
            order_id=order_id,
            blk_notify=True,
        )

    def _save_invoice_error_processing(
        self,               # type: SatProcessor
        pos_id,             # type: str
        order_id,           # type: str
        request,            # type: eTree.Element
        sat_service_name,   # type: str
    ):
        # type: (...) -> None

        dir_error = self._get_error_directory()

        file_name = "{0}_request_{1}_pos{2}_order{3}_erro.xml".format(
            order_id.zfill(9),
            sat_service_name,
            str(pos_id).zfill(2),
            order_id,
        )
        file_path = os.path.join(
            dir_error,
            file_name,
        )
        request_error = open(
            file_path,
            "w+",
        )
        request_error.write(str(request))
        request_error.close()

    def _get_error_directory(
        self,   # type: SatProcessor
    ):
        # type: (...) -> str

        dir_error = os.path.join(
            self.fiscal_sent_dir,
            "Erros",
        )
        if not os.path.exists(dir_error):
            os.makedirs(dir_error)

        return dir_error

    def _get_sent_directory(
        self,   # type: SatProcessor
    ):
        # type: (...) -> str

        dir_sent = os.path.join(
            self.fiscal_sent_dir,
            "Enviados",
        )
        if not os.path.exists(dir_sent):
            os.makedirs(dir_sent)

        return dir_sent

    @staticmethod
    def _get_invoice_directory(
        date_emission,  # type: str
        dir_sent,       # type: str
    ):
        # type: (...) -> str

        dir_nota = os.path.join(
            dir_sent,
            date_emission[0:4],
            date_emission[4:6],
            date_emission[6:8],
        )
        if not os.path.exists(dir_nota):
            os.makedirs(dir_nota)

        return dir_nota



# vkn.py - functions for handling the Turkish tax identification number
# coding: utf-8
#
# Copyright (C) 2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""VKN (Vergi Kimlik Numarası, Turkish tax identification number).

The Vergi Kimlik Numarası is the Turkish tax identification number used for
businesses. The number consists of 10 digits where the first digit is derived
from the company name.

More information:

* https://www.turkiye.gov.tr/gib-intvrg-vergi-kimlik-numarasi-dogrulama

>>> validate('4540536920')
'4540536920'
>>> validate('4540536921')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('454053692')
Traceback (most recent call last):
    ...
InvalidLength: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number).strip()


def calc_check_digit(number):
    """Calculate the check digit for the specified number."""
    s = 0
    for i, n in enumerate(reversed(number[:9]), 1):
        c1 = (int(n) + i) % 10
        if c1:
            c2 = (c1 * (2 ** i)) % 9 or 9
            s += c2
    return str((10 - s) % 10)


def validate(number):
    """Check if the number is a valid Vergi Kimlik Numarası. This checks the
    length and check digits."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 10:
        raise InvalidLength()
    if calc_check_digit(number) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid Vergi Kimlik Numarası. This checks the
    length and check digits."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from logging import Logger

from _OrderChangerProductionBox import OrderChangerProductionBox
from production.box import ProductionBox
from production.repository import ProductRepository
from typing import Optional, Union, List


class MasterPartCodeConverterBox(OrderChangerProductionBox):

    def __init__(self, name, sons, product_repository, logger=None):
        # type: (str, Optional[Union[List[ProductionBox]], ProductionBox], ProductRepository, Logger) -> None
        super(MasterPartCodeConverterBox, self).__init__(name, sons, logger)
        self.master_product_map = product_repository.get_master_product_map()

    def change_order(self, order):
        for item in order.items:
            if item.part_code in self.master_product_map:
                new_data = self.master_product_map[item.part_code]
                item.original_part_code = item.part_code
                item.original_description = item.description
                item.part_code = new_data[0]
                item.description = new_data[1]

        return order



from logging import Logger

from production.box import ProductionBox
from production.command import ChangePathCommand
from production.model import ProdStates
from typing import List, Tuple


class DistributorBox(ProductionBox):
    def __init__(self, name, paths, logger=None):
        # type: (str, List[Tuple[str, ProductionBox]], Logger) -> None
        self.path_box = {}
        self.paths = []
        self.paths_dict = {}
        self.active_path = None
        self.old_active = None

        sons = []
        index = 0
        for path in paths:
            if self.active_path is None:
                self.active_path = path

            path_obj = Path(path, index, True)

            self.paths.append(path_obj)
            self.paths_dict[path] = path_obj

            self.path_box[path] = index
            index += 1

        super(DistributorBox, self).__init__(name, sons, logger)

    def end_refresh(self):
        self.old_active = None
        super(DistributorBox, self).end_refresh()

    def order_modified(self, order):
        if self.old_active is not None:
            invalid_order = order.clone()
            invalid_order.prod_state = ProdStates.INVALID
            if hasattr(order, "stamped"):
                del invalid_order.stamped
            self.debug("DistributorBox {}: Invalidating order {} - path {}", self.name, order.order_id, self.old_active)
            self.sons[self.paths_dict[self.old_active].son_index].order_modified(invalid_order)
        if self.active_path is not None:
            self.debug("DistributorBox {}: Sending to {}", self.name, self.active_path)
            self.sons[self.paths_dict[self.active_path].son_index].order_modified(order)
        else:
            self.debug("DistributorBox {}: No active path.", self.name)

    def handle_production_command(self, command):
        if not isinstance(command, ChangePathCommand):
            return False

        result = False
        if command.path in self.paths_dict and self.paths_dict[command.path].enabled != command.enabled:
            self.paths_dict[command.path].enabled = command.enabled
            result = True

            self.old_active = self.active_path
            self.active_path = None
            for path in self.paths:
                if path.enabled:
                    self.active_path = path.name
                    break
            if self.active_path == self.old_active:
                self.old_active = None

        super(DistributorBox, self).handle_production_command(command)
        return result


class Path(object):
    def __init__(self, name, son_index, enabled):
        self.name = name
        self.son_index = son_index
        self.enabled = enabled



import json
from typing import Optional

from bustoken import (
    TK_NDISCOUNT_ON_TABLE_REOPEN,
    TK_NDISCOUNT_ON_TABLE_OPEN,
    TK_NDISCOUNT_ON_TAB_OPEN,
)
from msgbus import (
    FM_PARAM,
    MBException,
    MBEasyContext,
    TK_SYS_ACK,
)
from systools import sys_log_info


class DiscountService(object):
    @staticmethod
    def on_service_open(
        mbcontext,      # type: MBEasyContext
        pos_id,         # type: int
        table_id,       # type: str
        is_tab=False,   # type: Optional[bool]
    ):
        # type: (...) -> int

        token = TK_NDISCOUNT_ON_TAB_OPEN if is_tab else TK_NDISCOUNT_ON_TABLE_OPEN

        data = json.dumps(dict(
            pos_id=pos_id,
            table_id=table_id,
        ))
        try:
            msg = mbcontext.MB_EasySendMessage(
                dest_name="nDiscount",
                token=token,
                format=FM_PARAM,
                data=data,
                timeout=30000000,
            )

            if msg.token == TK_SYS_ACK:
                order_id = msg.data
                return order_id

        except MBException as ex:
            sys_log_info(message="Error on_service_open: {0}".format(ex.message))

    @staticmethod
    def on_reopen_service(
        mbcontext,  # type: MBEasyContext
        pos_id,     # type: int
        table_id,   # type: int
    ):
        # type: (...) -> Optional[bool]

        data = json.dumps(dict(
            pos_id=pos_id,
            table_id=table_id,
        ))
        try:
            mbcontext.MB_EasySendMessage(
                dest_name="nDiscount",
                token=TK_NDISCOUNT_ON_TABLE_REOPEN,
                format=FM_PARAM,
                data=data,
                timeout=30000000,
            )
        except MBException as ex:
            sys_log_info(message="Error on_reopen_service: {0}".format(ex.message))
            return False






from .core import *
from .codec import *

def ToASCII(label):
    return encode(label)

def ToUnicode(label):
    return decode(label)

def nameprep(s):
    raise NotImplementedError("IDNA 2008 does not utilise nameprep protocol")




# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class AsymmetricSignatureContext(object):
    @abc.abstractmethod
    def update(self, data):
        """
        Processes the provided bytes and returns nothing.
        """

    @abc.abstractmethod
    def finalize(self):
        """
        Returns the signature as bytes.
        """


@six.add_metaclass(abc.ABCMeta)
class AsymmetricVerificationContext(object):
    @abc.abstractmethod
    def update(self, data):
        """
        Processes the provided bytes and returns nothing.
        """

    @abc.abstractmethod
    def verify(self):
        """
        Raises an exception if the bytes provided to update do not match the
        signature or the signature does not match the public key.
        """



from abc import ABCMeta, abstractmethod

from customer_services.model import Customer


class CustomerRepository(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_customer(self, phone):
        # type: (str) -> Customer
    
        raise NotImplementedError

    @abstractmethod
    def add_customer(self, customer):
        # type: (Customer) -> None
    
        raise NotImplementedError

    @abstractmethod
    def update_customer(self, customer):
        # type: (Customer) -> None
    
        raise NotImplementedError



from __future__ import absolute_import
import collections
import functools
import logging

from ._collections import RecentlyUsedContainer
from .connectionpool import HTTPConnectionPool, HTTPSConnectionPool
from .connectionpool import port_by_scheme
from .exceptions import LocationValueError, MaxRetryError, ProxySchemeUnknown
from .packages.six.moves.urllib.parse import urljoin
from .request import RequestMethods
from .util.url import parse_url
from .util.retry import Retry


__all__ = ['PoolManager', 'ProxyManager', 'proxy_from_url']


log = logging.getLogger(__name__)

SSL_KEYWORDS = ('key_file', 'cert_file', 'cert_reqs', 'ca_certs',
                'ssl_version', 'ca_cert_dir', 'ssl_context')

# All known keyword arguments that could be provided to the pool manager, its
# pools, or the underlying connections. This is used to construct a pool key.
_key_fields = (
    'key_scheme',  # str
    'key_host',  # str
    'key_port',  # int
    'key_timeout',  # int or float or Timeout
    'key_retries',  # int or Retry
    'key_strict',  # bool
    'key_block',  # bool
    'key_source_address',  # str
    'key_key_file',  # str
    'key_cert_file',  # str
    'key_cert_reqs',  # str
    'key_ca_certs',  # str
    'key_ssl_version',  # str
    'key_ca_cert_dir',  # str
    'key_ssl_context',  # instance of ssl.SSLContext or urllib3.util.ssl_.SSLContext
    'key_maxsize',  # int
    'key_headers',  # dict
    'key__proxy',  # parsed proxy url
    'key__proxy_headers',  # dict
    'key_socket_options',  # list of (level (int), optname (int), value (int or str)) tuples
    'key__socks_options',  # dict
    'key_assert_hostname',  # bool or string
    'key_assert_fingerprint',  # str
)

#: The namedtuple class used to construct keys for the connection pool.
#: All custom key schemes should include the fields in this key at a minimum.
PoolKey = collections.namedtuple('PoolKey', _key_fields)


def _default_key_normalizer(key_class, request_context):
    """
    Create a pool key out of a request context dictionary.

    According to RFC 3986, both the scheme and host are case-insensitive.
    Therefore, this function normalizes both before constructing the pool
    key for an HTTPS request. If you wish to change this behaviour, provide
    alternate callables to ``key_fn_by_scheme``.

    :param key_class:
        The class to use when constructing the key. This should be a namedtuple
        with the ``scheme`` and ``host`` keys at a minimum.
    :type  key_class: namedtuple
    :param request_context:
        A dictionary-like object that contain the context for a request.
    :type  request_context: dict

    :return: A namedtuple that can be used as a connection pool key.
    :rtype:  PoolKey
    """
    # Since we mutate the dictionary, make a copy first
    context = request_context.copy()
    context['scheme'] = context['scheme'].lower()
    context['host'] = context['host'].lower()

    # These are both dictionaries and need to be transformed into frozensets
    for key in ('headers', '_proxy_headers', '_socks_options'):
        if key in context and context[key] is not None:
            context[key] = frozenset(context[key].items())

    # The socket_options key may be a list and needs to be transformed into a
    # tuple.
    socket_opts = context.get('socket_options')
    if socket_opts is not None:
        context['socket_options'] = tuple(socket_opts)

    # Map the kwargs to the names in the namedtuple - this is necessary since
    # namedtuples can't have fields starting with '_'.
    for key in list(context.keys()):
        context['key_' + key] = context.pop(key)

    # Default to ``None`` for keys missing from the context
    for field in key_class._fields:
        if field not in context:
            context[field] = None

    return key_class(**context)


#: A dictionary that maps a scheme to a callable that creates a pool key.
#: This can be used to alter the way pool keys are constructed, if desired.
#: Each PoolManager makes a copy of this dictionary so they can be configured
#: globally here, or individually on the instance.
key_fn_by_scheme = {
    'http': functools.partial(_default_key_normalizer, PoolKey),
    'https': functools.partial(_default_key_normalizer, PoolKey),
}

pool_classes_by_scheme = {
    'http': HTTPConnectionPool,
    'https': HTTPSConnectionPool,
}


class PoolManager(RequestMethods):
    """
    Allows for arbitrary requests while transparently keeping track of
    necessary connection pools for you.

    :param num_pools:
        Number of connection pools to cache before discarding the least
        recently used pool.

    :param headers:
        Headers to include with all requests, unless other headers are given
        explicitly.

    :param \\**connection_pool_kw:
        Additional parameters are used to create fresh
        :class:`urllib3.connectionpool.ConnectionPool` instances.

    Example::

        >>> manager = PoolManager(num_pools=2)
        >>> r = manager.request('GET', 'http://google.com/')
        >>> r = manager.request('GET', 'http://google.com/mail')
        >>> r = manager.request('GET', 'http://yahoo.com/')
        >>> len(manager.pools)
        2

    """

    proxy = None

    def __init__(self, num_pools=10, headers=None, **connection_pool_kw):
        RequestMethods.__init__(self, headers)
        self.connection_pool_kw = connection_pool_kw
        self.pools = RecentlyUsedContainer(num_pools,
                                           dispose_func=lambda p: p.close())

        # Locally set the pool classes and keys so other PoolManagers can
        # override them.
        self.pool_classes_by_scheme = pool_classes_by_scheme
        self.key_fn_by_scheme = key_fn_by_scheme.copy()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.clear()
        # Return False to re-raise any potential exceptions
        return False

    def _new_pool(self, scheme, host, port, request_context=None):
        """
        Create a new :class:`ConnectionPool` based on host, port, scheme, and
        any additional pool keyword arguments.

        If ``request_context`` is provided, it is provided as keyword arguments
        to the pool class used. This method is used to actually create the
        connection pools handed out by :meth:`connection_from_url` and
        companion methods. It is intended to be overridden for customization.
        """
        pool_cls = self.pool_classes_by_scheme[scheme]
        if request_context is None:
            request_context = self.connection_pool_kw.copy()

        # Although the context has everything necessary to create the pool,
        # this function has historically only used the scheme, host, and port
        # in the positional args. When an API change is acceptable these can
        # be removed.
        for key in ('scheme', 'host', 'port'):
            request_context.pop(key, None)

        if scheme == 'http':
            for kw in SSL_KEYWORDS:
                request_context.pop(kw, None)

        return pool_cls(host, port, **request_context)

    def clear(self):
        """
        Empty our store of pools and direct them all to close.

        This will not affect in-flight connections, but they will not be
        re-used after completion.
        """
        self.pools.clear()

    def connection_from_host(self, host, port=None, scheme='http', pool_kwargs=None):
        """
        Get a :class:`ConnectionPool` based on the host, port, and scheme.

        If ``port`` isn't given, it will be derived from the ``scheme`` using
        ``urllib3.connectionpool.port_by_scheme``. If ``pool_kwargs`` is
        provided, it is merged with the instance's ``connection_pool_kw``
        variable and used to create the new connection pool, if one is
        needed.
        """

        if not host:
            raise LocationValueError("No host specified.")

        request_context = self._merge_pool_kwargs(pool_kwargs)
        request_context['scheme'] = scheme or 'http'
        if not port:
            port = port_by_scheme.get(request_context['scheme'].lower(), 80)
        request_context['port'] = port
        request_context['host'] = host

        return self.connection_from_context(request_context)

    def connection_from_context(self, request_context):
        """
        Get a :class:`ConnectionPool` based on the request context.

        ``request_context`` must at least contain the ``scheme`` key and its
        value must be a key in ``key_fn_by_scheme`` instance variable.
        """
        scheme = request_context['scheme'].lower()
        pool_key_constructor = self.key_fn_by_scheme[scheme]
        pool_key = pool_key_constructor(request_context)

        return self.connection_from_pool_key(pool_key, request_context=request_context)

    def connection_from_pool_key(self, pool_key, request_context=None):
        """
        Get a :class:`ConnectionPool` based on the provided pool key.

        ``pool_key`` should be a namedtuple that only contains immutable
        objects. At a minimum it must have the ``scheme``, ``host``, and
        ``port`` fields.
        """
        with self.pools.lock:
            # If the scheme, host, or port doesn't match existing open
            # connections, open a new ConnectionPool.
            pool = self.pools.get(pool_key)
            if pool:
                return pool

            # Make a fresh ConnectionPool of the desired type
            scheme = request_context['scheme']
            host = request_context['host']
            port = request_context['port']
            pool = self._new_pool(scheme, host, port, request_context=request_context)
            self.pools[pool_key] = pool

        return pool

    def connection_from_url(self, url, pool_kwargs=None):
        """
        Similar to :func:`urllib3.connectionpool.connection_from_url`.

        If ``pool_kwargs`` is not provided and a new pool needs to be
        constructed, ``self.connection_pool_kw`` is used to initialize
        the :class:`urllib3.connectionpool.ConnectionPool`. If ``pool_kwargs``
        is provided, it is used instead. Note that if a new pool does not
        need to be created for the request, the provided ``pool_kwargs`` are
        not used.
        """
        u = parse_url(url)
        return self.connection_from_host(u.host, port=u.port, scheme=u.scheme,
                                         pool_kwargs=pool_kwargs)

    def _merge_pool_kwargs(self, override):
        """
        Merge a dictionary of override values for self.connection_pool_kw.

        This does not modify self.connection_pool_kw and returns a new dict.
        Any keys in the override dictionary with a value of ``None`` are
        removed from the merged dictionary.
        """
        base_pool_kwargs = self.connection_pool_kw.copy()
        if override:
            for key, value in override.items():
                if value is None:
                    try:
                        del base_pool_kwargs[key]
                    except KeyError:
                        pass
                else:
                    base_pool_kwargs[key] = value
        return base_pool_kwargs

    def urlopen(self, method, url, redirect=True, **kw):
        """
        Same as :meth:`urllib3.connectionpool.HTTPConnectionPool.urlopen`
        with custom cross-host redirect logic and only sends the request-uri
        portion of the ``url``.

        The given ``url`` parameter must be absolute, such that an appropriate
        :class:`urllib3.connectionpool.ConnectionPool` can be chosen for it.
        """
        u = parse_url(url)
        conn = self.connection_from_host(u.host, port=u.port, scheme=u.scheme)

        kw['assert_same_host'] = False
        kw['redirect'] = False

        if 'headers' not in kw:
            kw['headers'] = self.headers.copy()

        if self.proxy is not None and u.scheme == "http":
            response = conn.urlopen(method, url, **kw)
        else:
            response = conn.urlopen(method, u.request_uri, **kw)

        redirect_location = redirect and response.get_redirect_location()
        if not redirect_location:
            return response

        # Support relative URLs for redirecting.
        redirect_location = urljoin(url, redirect_location)

        # RFC 7231, Section 6.4.4
        if response.status == 303:
            method = 'GET'

        retries = kw.get('retries')
        if not isinstance(retries, Retry):
            retries = Retry.from_int(retries, redirect=redirect)

        # Strip headers marked as unsafe to forward to the redirected location.
        # Check remove_headers_on_redirect to avoid a potential network call within
        # conn.is_same_host() which may use socket.gethostbyname() in the future.
        if (retries.remove_headers_on_redirect
                and not conn.is_same_host(redirect_location)):
            for header in retries.remove_headers_on_redirect:
                kw['headers'].pop(header, None)

        try:
            retries = retries.increment(method, url, response=response, _pool=conn)
        except MaxRetryError:
            if retries.raise_on_redirect:
                raise
            return response

        kw['retries'] = retries
        kw['redirect'] = redirect

        log.info("Redirecting %s -> %s", url, redirect_location)
        return self.urlopen(method, redirect_location, **kw)


class ProxyManager(PoolManager):
    """
    Behaves just like :class:`PoolManager`, but sends all requests through
    the defined proxy, using the CONNECT method for HTTPS URLs.

    :param proxy_url:
        The URL of the proxy to be used.

    :param proxy_headers:
        A dictionary containing headers that will be sent to the proxy. In case
        of HTTP they are being sent with each request, while in the
        HTTPS/CONNECT case they are sent only once. Could be used for proxy
        authentication.

    Example:
        >>> proxy = urllib3.ProxyManager('http://localhost:3128/')
        >>> r1 = proxy.request('GET', 'http://google.com/')
        >>> r2 = proxy.request('GET', 'http://httpbin.org/')
        >>> len(proxy.pools)
        1
        >>> r3 = proxy.request('GET', 'https://httpbin.org/')
        >>> r4 = proxy.request('GET', 'https://twitter.com/')
        >>> len(proxy.pools)
        3

    """

    def __init__(self, proxy_url, num_pools=10, headers=None,
                 proxy_headers=None, **connection_pool_kw):

        if isinstance(proxy_url, HTTPConnectionPool):
            proxy_url = '%s://%s:%i' % (proxy_url.scheme, proxy_url.host,
                                        proxy_url.port)
        proxy = parse_url(proxy_url)
        if not proxy.port:
            port = port_by_scheme.get(proxy.scheme, 80)
            proxy = proxy._replace(port=port)

        if proxy.scheme not in ("http", "https"):
            raise ProxySchemeUnknown(proxy.scheme)

        self.proxy = proxy
        self.proxy_headers = proxy_headers or {}

        connection_pool_kw['_proxy'] = self.proxy
        connection_pool_kw['_proxy_headers'] = self.proxy_headers

        super(ProxyManager, self).__init__(
            num_pools, headers, **connection_pool_kw)

    def connection_from_host(self, host, port=None, scheme='http', pool_kwargs=None):
        if scheme == "https":
            return super(ProxyManager, self).connection_from_host(
                host, port, scheme, pool_kwargs=pool_kwargs)

        return super(ProxyManager, self).connection_from_host(
            self.proxy.host, self.proxy.port, self.proxy.scheme, pool_kwargs=pool_kwargs)

    def _set_proxy_headers(self, url, headers=None):
        """
        Sets headers needed by proxies: specifically, the Accept and Host
        headers. Only sets headers not provided by the user.
        """
        headers_ = {'Accept': '*/*'}

        netloc = parse_url(url).netloc
        if netloc:
            headers_['Host'] = netloc

        if headers:
            headers_.update(headers)
        return headers_

    def urlopen(self, method, url, redirect=True, **kw):
        "Same as HTTP(S)ConnectionPool.urlopen, ``url`` must be absolute."
        u = parse_url(url)

        if u.scheme == "http":
            # For proxied HTTPS requests, httplib sets the necessary headers
            # on the CONNECT to the proxy. For HTTP, we'll definitely
            # need to set 'Host' at the very least.
            headers = kw.get('headers', self.headers)
            kw['headers'] = self._set_proxy_headers(url, headers)

        return super(ProxyManager, self).urlopen(method, url, redirect=redirect, **kw)


def proxy_from_url(url, **kw):
    return ProxyManager(proxy_url=url, **kw)



# -*- coding: utf-8 -*-
import datetime

import msgbus
import sysactions

from actions.goals import update_daily_goals
from actions.login.open_operator_min_level import get_min_level_to_open_operator
from actions.util import get_operator_info
from utils import (
    open_user,
    login_user,
    print_open_day_report,
    get_pod_type_and_pos_function,
    change_pod_type_and_pos_function,
    close_user,
    get_initial_float,
    save_initial_float,
    user_inserted_initial_amount,
    get_authorization_if_necessary,
    save_last_authorization,
)

from .. import mb_context
from ..drawer.do_open_drawer import doOpenDrawer

USECS_PER_SEC = 1000000


def sign_in_quick_user(pos_id, user_info):
    model = sysactions.get_model(pos_id)
    min_level = get_min_level_to_open_operator("QSR")
    authorization_response = get_authorization_if_necessary(model, pos_id, user_info, min_level)
    if not authorization_response.success:
        return False

    user_id = user_info.get("UserId")
    pod_type, pos_function = get_pod_type_and_pos_function(pos_id)
    if pod_type is None or pos_function is None:
        return False

    initial_float = 0
    _, session = get_operator_info(model, user_id)
    if not user_inserted_initial_amount(pos_id, session):
        initial_float = get_initial_float(pos_id, pod_type, pos_function)

    if open_user(pos_id, user_id) is False:
        return False

    if login_user(pos_id, user_id) is False:
        close_user(pos_id, user_id)
        return False

    save_last_authorization(pos_id, authorization_response.authorized_by, authorization_response.authorized_by_level)

    change_pod_type_and_pos_function(pos_id, pod_type, pos_function)
    save_initial_float(initial_float, pos_id, user_id)
    update_daily_goals(
        pos_id=pos_id,
        synchronous=False,
    )

    if int(initial_float) > 0:
        doOpenDrawer(pos_id)

    print_open_day_report(pos_id, initial_float)

    return True


def must_change_business_period(model):
    current_date = datetime.datetime.now().strftime("%Y%m%d")
    return model.find("PosState").get("period") != current_date or model.find("PosState").get("state") != "OPENED"


def must_close_business_period(model):
    return model.find("PosState").get("state") == "OPENED" or model.find("PosState").get("state") == "BLOCKED"


def open_business_period(pos_id):
    current_date = datetime.datetime.now().strftime("%Y%m%d")
    business_begin_msg = mb_context.MB_EasySendMessage(
        "POS{}".format(pos_id),
        msgbus.TK_POS_BUSINESSBEGIN,
        msgbus.FM_PARAM,
        "{}\0{}".format(pos_id, current_date),
        timeout=90000000
    )

    if business_begin_msg.token != msgbus.TK_SYS_ACK:
        if business_begin_msg.data:
            raise Exception("Error opening day: {}".format(business_begin_msg.data))

        raise Exception("Error opening day")


def get_pos_service(pos_id):
    return "POS%d" % int(pos_id)



from datetime import datetime


class UpdateData(object):
    def __init__(self,
                 update_id=None,
                 update_name=None,
                 obtained_date=None,
                 downloaded_date=None,
                 backup_date=None,
                 applied_date=None,
                 notified_date=None):
        # type: (int, str, datetime, datetime, datetime, datetime, datetime) -> None

        self.update_id = update_id
        self.update_name = update_name
        self.obtained_date = obtained_date
        self.downloaded_date = downloaded_date
        self.backup_date = backup_date
        self.applied_date = applied_date
        self.notified_date = notified_date



#! python
#
# This module implements a RFC2217 compatible client. RF2217 descibes a
# protocol to access serial ports over TCP/IP and allows setting the baud rate,
# modem control lines etc.
#
# This file is part of pySerial. https://github.com/pyserial/pyserial
# (C) 2001-2015 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause

# TODO:
# - setting control line -> answer is not checked (had problems with one of the
#   severs). consider implementing a compatibility mode flag to make check
#   conditional
# - write timeout not implemented at all

# ###########################################################################
# observations and issues with servers
# ===========================================================================
# sredird V2.2.1
# - http://www.ibiblio.org/pub/Linux/system/serial/   sredird-2.2.2.tar.gz
# - does not acknowledge SET_CONTROL (RTS/DTR) correctly, always responding
#   [105 1] instead of the actual value.
# - SET_BAUDRATE answer contains 4 extra null bytes -> probably for larger
#   numbers than 2**32?
# - To get the signature [COM_PORT_OPTION 0] has to be sent.
# - run a server: while true; do nc -l -p 7000 -c "sredird debug /dev/ttyUSB0 /var/lock/sredir"; done
# ===========================================================================
# telnetcpcd (untested)
# - http://ftp.wayne.edu/kermit/sredird/telnetcpcd-1.09.tar.gz
# - To get the signature [COM_PORT_OPTION] w/o data has to be sent.
# ===========================================================================
# ser2net
# - does not negotiate BINARY or COM_PORT_OPTION for his side but at least
#   acknowledges that the client activates these options
# - The configuration may be that the server prints a banner. As this client
#   implementation does a flushInput on connect, this banner is hidden from
#   the user application.
# - NOTIFY_MODEMSTATE: the poll interval of the server seems to be one
#   second.
# - To get the signature [COM_PORT_OPTION 0] has to be sent.
# - run a server: run ser2net daemon, in /etc/ser2net.conf:
#     2000:telnet:0:/dev/ttyS0:9600 remctl banner
# ###########################################################################

# How to identify ports? pySerial might want to support other protocols in the
# future, so lets use an URL scheme.
# for RFC2217 compliant servers we will use this:
#    rfc2217://<host>:<port>[?option[&option...]]
#
# options:
# - "logging" set log level print diagnostic messages (e.g. "logging=debug")
# - "ign_set_control": do not look at the answers to SET_CONTROL
# - "poll_modem": issue NOTIFY_MODEMSTATE requests when CTS/DTR/RI/CD is read.
#   Without this option it expects that the server sends notifications
#   automatically on change (which most servers do and is according to the
#   RFC).
# the order of the options is not relevant

from __future__ import absolute_import

import logging
import socket
import struct
import threading
import time
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse
try:
    import Queue
except ImportError:
    import queue as Queue

import serial
from serial.serialutil import SerialBase, SerialException, to_bytes, \
    iterbytes, PortNotOpenError, Timeout

# port string is expected to be something like this:
# rfc2217://host:port
# host may be an IP or including domain, whatever.
# port is 0...65535

# map log level names to constants. used in from_url()
LOGGER_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
}


# telnet protocol characters
SE = b'\xf0'    # Subnegotiation End
NOP = b'\xf1'   # No Operation
DM = b'\xf2'    # Data Mark
BRK = b'\xf3'   # Break
IP = b'\xf4'    # Interrupt process
AO = b'\xf5'    # Abort output
AYT = b'\xf6'   # Are You There
EC = b'\xf7'    # Erase Character
EL = b'\xf8'    # Erase Line
GA = b'\xf9'    # Go Ahead
SB = b'\xfa'    # Subnegotiation Begin
WILL = b'\xfb'
WONT = b'\xfc'
DO = b'\xfd'
DONT = b'\xfe'
IAC = b'\xff'   # Interpret As Command
IAC_DOUBLED = b'\xff\xff'

# selected telnet options
BINARY = b'\x00'    # 8-bit data path
ECHO = b'\x01'      # echo
SGA = b'\x03'       # suppress go ahead

# RFC2217
COM_PORT_OPTION = b'\x2c'

# Client to Access Server
SET_BAUDRATE = b'\x01'
SET_DATASIZE = b'\x02'
SET_PARITY = b'\x03'
SET_STOPSIZE = b'\x04'
SET_CONTROL = b'\x05'
NOTIFY_LINESTATE = b'\x06'
NOTIFY_MODEMSTATE = b'\x07'
FLOWCONTROL_SUSPEND = b'\x08'
FLOWCONTROL_RESUME = b'\x09'
SET_LINESTATE_MASK = b'\x0a'
SET_MODEMSTATE_MASK = b'\x0b'
PURGE_DATA = b'\x0c'

SERVER_SET_BAUDRATE = b'\x65'
SERVER_SET_DATASIZE = b'\x66'
SERVER_SET_PARITY = b'\x67'
SERVER_SET_STOPSIZE = b'\x68'
SERVER_SET_CONTROL = b'\x69'
SERVER_NOTIFY_LINESTATE = b'\x6a'
SERVER_NOTIFY_MODEMSTATE = b'\x6b'
SERVER_FLOWCONTROL_SUSPEND = b'\x6c'
SERVER_FLOWCONTROL_RESUME = b'\x6d'
SERVER_SET_LINESTATE_MASK = b'\x6e'
SERVER_SET_MODEMSTATE_MASK = b'\x6f'
SERVER_PURGE_DATA = b'\x70'

RFC2217_ANSWER_MAP = {
    SET_BAUDRATE: SERVER_SET_BAUDRATE,
    SET_DATASIZE: SERVER_SET_DATASIZE,
    SET_PARITY: SERVER_SET_PARITY,
    SET_STOPSIZE: SERVER_SET_STOPSIZE,
    SET_CONTROL: SERVER_SET_CONTROL,
    NOTIFY_LINESTATE: SERVER_NOTIFY_LINESTATE,
    NOTIFY_MODEMSTATE: SERVER_NOTIFY_MODEMSTATE,
    FLOWCONTROL_SUSPEND: SERVER_FLOWCONTROL_SUSPEND,
    FLOWCONTROL_RESUME: SERVER_FLOWCONTROL_RESUME,
    SET_LINESTATE_MASK: SERVER_SET_LINESTATE_MASK,
    SET_MODEMSTATE_MASK: SERVER_SET_MODEMSTATE_MASK,
    PURGE_DATA: SERVER_PURGE_DATA,
}

SET_CONTROL_REQ_FLOW_SETTING = b'\x00'        # Request Com Port Flow Control Setting (outbound/both)
SET_CONTROL_USE_NO_FLOW_CONTROL = b'\x01'     # Use No Flow Control (outbound/both)
SET_CONTROL_USE_SW_FLOW_CONTROL = b'\x02'     # Use XON/XOFF Flow Control (outbound/both)
SET_CONTROL_USE_HW_FLOW_CONTROL = b'\x03'     # Use HARDWARE Flow Control (outbound/both)
SET_CONTROL_REQ_BREAK_STATE = b'\x04'         # Request BREAK State
SET_CONTROL_BREAK_ON = b'\x05'                # Set BREAK State ON
SET_CONTROL_BREAK_OFF = b'\x06'               # Set BREAK State OFF
SET_CONTROL_REQ_DTR = b'\x07'                 # Request DTR Signal State
SET_CONTROL_DTR_ON = b'\x08'                  # Set DTR Signal State ON
SET_CONTROL_DTR_OFF = b'\x09'                 # Set DTR Signal State OFF
SET_CONTROL_REQ_RTS = b'\x0a'                 # Request RTS Signal State
SET_CONTROL_RTS_ON = b'\x0b'                  # Set RTS Signal State ON
SET_CONTROL_RTS_OFF = b'\x0c'                 # Set RTS Signal State OFF
SET_CONTROL_REQ_FLOW_SETTING_IN = b'\x0d'     # Request Com Port Flow Control Setting (inbound)
SET_CONTROL_USE_NO_FLOW_CONTROL_IN = b'\x0e'  # Use No Flow Control (inbound)
SET_CONTROL_USE_SW_FLOW_CONTOL_IN = b'\x0f'   # Use XON/XOFF Flow Control (inbound)
SET_CONTROL_USE_HW_FLOW_CONTOL_IN = b'\x10'   # Use HARDWARE Flow Control (inbound)
SET_CONTROL_USE_DCD_FLOW_CONTROL = b'\x11'    # Use DCD Flow Control (outbound/both)
SET_CONTROL_USE_DTR_FLOW_CONTROL = b'\x12'    # Use DTR Flow Control (inbound)
SET_CONTROL_USE_DSR_FLOW_CONTROL = b'\x13'    # Use DSR Flow Control (outbound/both)

LINESTATE_MASK_TIMEOUT = 128        # Time-out Error
LINESTATE_MASK_SHIFTREG_EMPTY = 64  # Transfer Shift Register Empty
LINESTATE_MASK_TRANSREG_EMPTY = 32  # Transfer Holding Register Empty
LINESTATE_MASK_BREAK_DETECT = 16    # Break-detect Error
LINESTATE_MASK_FRAMING_ERROR = 8    # Framing Error
LINESTATE_MASK_PARTIY_ERROR = 4     # Parity Error
LINESTATE_MASK_OVERRUN_ERROR = 2    # Overrun Error
LINESTATE_MASK_DATA_READY = 1       # Data Ready

MODEMSTATE_MASK_CD = 128            # Receive Line Signal Detect (also known as Carrier Detect)
MODEMSTATE_MASK_RI = 64             # Ring Indicator
MODEMSTATE_MASK_DSR = 32            # Data-Set-Ready Signal State
MODEMSTATE_MASK_CTS = 16            # Clear-To-Send Signal State
MODEMSTATE_MASK_CD_CHANGE = 8       # Delta Receive Line Signal Detect
MODEMSTATE_MASK_RI_CHANGE = 4       # Trailing-edge Ring Detector
MODEMSTATE_MASK_DSR_CHANGE = 2      # Delta Data-Set-Ready
MODEMSTATE_MASK_CTS_CHANGE = 1      # Delta Clear-To-Send

PURGE_RECEIVE_BUFFER = b'\x01'      # Purge access server receive data buffer
PURGE_TRANSMIT_BUFFER = b'\x02'     # Purge access server transmit data buffer
PURGE_BOTH_BUFFERS = b'\x03'        # Purge both the access server receive data
                                    # buffer and the access server transmit data buffer


RFC2217_PARITY_MAP = {
    serial.PARITY_NONE: 1,
    serial.PARITY_ODD: 2,
    serial.PARITY_EVEN: 3,
    serial.PARITY_MARK: 4,
    serial.PARITY_SPACE: 5,
}
RFC2217_REVERSE_PARITY_MAP = dict((v, k) for k, v in RFC2217_PARITY_MAP.items())

RFC2217_STOPBIT_MAP = {
    serial.STOPBITS_ONE: 1,
    serial.STOPBITS_ONE_POINT_FIVE: 3,
    serial.STOPBITS_TWO: 2,
}
RFC2217_REVERSE_STOPBIT_MAP = dict((v, k) for k, v in RFC2217_STOPBIT_MAP.items())

# Telnet filter states
M_NORMAL = 0
M_IAC_SEEN = 1
M_NEGOTIATE = 2

# TelnetOption and TelnetSubnegotiation states
REQUESTED = 'REQUESTED'
ACTIVE = 'ACTIVE'
INACTIVE = 'INACTIVE'
REALLY_INACTIVE = 'REALLY_INACTIVE'


class TelnetOption(object):
    """Manage a single telnet option, keeps track of DO/DONT WILL/WONT."""

    def __init__(self, connection, name, option, send_yes, send_no, ack_yes,
                 ack_no, initial_state, activation_callback=None):
        """\
        Initialize option.
        :param connection: connection used to transmit answers
        :param name: a readable name for debug outputs
        :param send_yes: what to send when option is to be enabled.
        :param send_no: what to send when option is to be disabled.
        :param ack_yes: what to expect when remote agrees on option.
        :param ack_no: what to expect when remote disagrees on option.
        :param initial_state: options initialized with REQUESTED are tried to
            be enabled on startup. use INACTIVE for all others.
        """
        self.connection = connection
        self.name = name
        self.option = option
        self.send_yes = send_yes
        self.send_no = send_no
        self.ack_yes = ack_yes
        self.ack_no = ack_no
        self.state = initial_state
        self.active = False
        self.activation_callback = activation_callback

    def __repr__(self):
        """String for debug outputs"""
        return "{o.name}:{o.active}({o.state})".format(o=self)

    def process_incoming(self, command):
        """\
        A DO/DONT/WILL/WONT was received for this option, update state and
        answer when needed.
        """
        if command == self.ack_yes:
            if self.state is REQUESTED:
                self.state = ACTIVE
                self.active = True
                if self.activation_callback is not None:
                    self.activation_callback()
            elif self.state is ACTIVE:
                pass
            elif self.state is INACTIVE:
                self.state = ACTIVE
                self.connection.telnet_send_option(self.send_yes, self.option)
                self.active = True
                if self.activation_callback is not None:
                    self.activation_callback()
            elif self.state is REALLY_INACTIVE:
                self.connection.telnet_send_option(self.send_no, self.option)
            else:
                raise ValueError('option in illegal state {!r}'.format(self))
        elif command == self.ack_no:
            if self.state is REQUESTED:
                self.state = INACTIVE
                self.active = False
            elif self.state is ACTIVE:
                self.state = INACTIVE
                self.connection.telnet_send_option(self.send_no, self.option)
                self.active = False
            elif self.state is INACTIVE:
                pass
            elif self.state is REALLY_INACTIVE:
                pass
            else:
                raise ValueError('option in illegal state {!r}'.format(self))


class TelnetSubnegotiation(object):
    """\
    A object to handle subnegotiation of options. In this case actually
    sub-sub options for RFC 2217. It is used to track com port options.
    """

    def __init__(self, connection, name, option, ack_option=None):
        if ack_option is None:
            ack_option = option
        self.connection = connection
        self.name = name
        self.option = option
        self.value = None
        self.ack_option = ack_option
        self.state = INACTIVE

    def __repr__(self):
        """String for debug outputs."""
        return "{sn.name}:{sn.state}".format(sn=self)

    def set(self, value):
        """\
        Request a change of the value. a request is sent to the server. if
        the client needs to know if the change is performed he has to check the
        state of this object.
        """
        self.value = value
        self.state = REQUESTED
        self.connection.rfc2217_send_subnegotiation(self.option, self.value)
        if self.connection.logger:
            self.connection.logger.debug("SB Requesting {} -> {!r}".format(self.name, self.value))

    def is_ready(self):
        """\
        Check if answer from server has been received. when server rejects
        the change, raise a ValueError.
        """
        if self.state == REALLY_INACTIVE:
            raise ValueError("remote rejected value for option {!r}".format(self.name))
        return self.state == ACTIVE
    # add property to have a similar interface as TelnetOption
    active = property(is_ready)

    def wait(self, timeout=3):
        """\
        Wait until the subnegotiation has been acknowledged or timeout. It
        can also throw a value error when the answer from the server does not
        match the value sent.
        """
        timeout_timer = Timeout(timeout)
        while not timeout_timer.expired():
            time.sleep(0.05)    # prevent 100% CPU load
            if self.is_ready():
                break
        else:
            raise SerialException("timeout while waiting for option {!r}".format(self.name))

    def check_answer(self, suboption):
        """\
        Check an incoming subnegotiation block. The parameter already has
        cut off the header like sub option number and com port option value.
        """
        if self.value == suboption[:len(self.value)]:
            self.state = ACTIVE
        else:
            # error propagation done in is_ready
            self.state = REALLY_INACTIVE
        if self.connection.logger:
            self.connection.logger.debug("SB Answer {} -> {!r} -> {}".format(self.name, suboption, self.state))


class Serial(SerialBase):
    """Serial port implementation for RFC 2217 remote serial ports."""

    BAUDRATES = (50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
                 9600, 19200, 38400, 57600, 115200)

    def __init__(self, *args, **kwargs):
        self._thread = None
        self._socket = None
        self._linestate = 0
        self._modemstate = None
        self._modemstate_timeout = Timeout(-1)
        self._remote_suspend_flow = False
        self._write_lock = None
        self.logger = None
        self._ignore_set_control_answer = False
        self._poll_modem_state = False
        self._network_timeout = 3
        self._telnet_options = None
        self._rfc2217_port_settings = None
        self._rfc2217_options = None
        self._read_buffer = None
        super(Serial, self).__init__(*args, **kwargs)  # must be last call in case of auto-open

    def open(self):
        """\
        Open port with current settings. This may throw a SerialException
        if the port cannot be opened.
        """
        self.logger = None
        self._ignore_set_control_answer = False
        self._poll_modem_state = False
        self._network_timeout = 3
        if self._port is None:
            raise SerialException("Port must be configured before it can be used.")
        if self.is_open:
            raise SerialException("Port is already open.")
        try:
            self._socket = socket.create_connection(self.from_url(self.portstr), timeout=5)  # XXX good value?
            self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        except Exception as msg:
            self._socket = None
            raise SerialException("Could not open port {}: {}".format(self.portstr, msg))

        # use a thread save queue as buffer. it also simplifies implementing
        # the read timeout
        self._read_buffer = Queue.Queue()
        # to ensure that user writes does not interfere with internal
        # telnet/rfc2217 options establish a lock
        self._write_lock = threading.Lock()
        # name the following separately so that, below, a check can be easily done
        mandadory_options = [
            TelnetOption(self, 'we-BINARY', BINARY, WILL, WONT, DO, DONT, INACTIVE),
            TelnetOption(self, 'we-RFC2217', COM_PORT_OPTION, WILL, WONT, DO, DONT, REQUESTED),
        ]
        # all supported telnet options
        self._telnet_options = [
            TelnetOption(self, 'ECHO', ECHO, DO, DONT, WILL, WONT, REQUESTED),
            TelnetOption(self, 'we-SGA', SGA, WILL, WONT, DO, DONT, REQUESTED),
            TelnetOption(self, 'they-SGA', SGA, DO, DONT, WILL, WONT, REQUESTED),
            TelnetOption(self, 'they-BINARY', BINARY, DO, DONT, WILL, WONT, INACTIVE),
            TelnetOption(self, 'they-RFC2217', COM_PORT_OPTION, DO, DONT, WILL, WONT, REQUESTED),
        ] + mandadory_options
        # RFC 2217 specific states
        # COM port settings
        self._rfc2217_port_settings = {
            'baudrate': TelnetSubnegotiation(self, 'baudrate', SET_BAUDRATE, SERVER_SET_BAUDRATE),
            'datasize': TelnetSubnegotiation(self, 'datasize', SET_DATASIZE, SERVER_SET_DATASIZE),
            'parity':   TelnetSubnegotiation(self, 'parity',   SET_PARITY,   SERVER_SET_PARITY),
            'stopsize': TelnetSubnegotiation(self, 'stopsize', SET_STOPSIZE, SERVER_SET_STOPSIZE),
        }
        # There are more subnegotiation objects, combine all in one dictionary
        # for easy access
        self._rfc2217_options = {
            'purge':    TelnetSubnegotiation(self, 'purge',    PURGE_DATA,   SERVER_PURGE_DATA),
            'control':  TelnetSubnegotiation(self, 'control',  SET_CONTROL,  SERVER_SET_CONTROL),
        }
        self._rfc2217_options.update(self._rfc2217_port_settings)
        # cache for line and modem states that the server sends to us
        self._linestate = 0
        self._modemstate = None
        self._modemstate_timeout = Timeout(-1)
        # RFC 2217 flow control between server and client
        self._remote_suspend_flow = False

        self.is_open = True
        self._thread = threading.Thread(target=self._telnet_read_loop)
        self._thread.setDaemon(True)
        self._thread.setName('pySerial RFC 2217 reader thread for {}'.format(self._port))
        self._thread.start()

        try:    # must clean-up if open fails
            # negotiate Telnet/RFC 2217 -> send initial requests
            for option in self._telnet_options:
                if option.state is REQUESTED:
                    self.telnet_send_option(option.send_yes, option.option)
            # now wait until important options are negotiated
            timeout = Timeout(self._network_timeout)
            while not timeout.expired():
                time.sleep(0.05)    # prevent 100% CPU load
                if sum(o.active for o in mandadory_options) == sum(o.state != INACTIVE for o in mandadory_options):
                    break
            else:
                raise SerialException(
                    "Remote does not seem to support RFC2217 or BINARY mode {!r}".format(mandadory_options))
            if self.logger:
                self.logger.info("Negotiated options: {}".format(self._telnet_options))

            # fine, go on, set RFC 2217 specific things
            self._reconfigure_port()
            # all things set up get, now a clean start
            if not self._dsrdtr:
                self._update_dtr_state()
            if not self._rtscts:
                self._update_rts_state()
            self.reset_input_buffer()
            self.reset_output_buffer()
        except:
            self.close()
            raise

    def _reconfigure_port(self):
        """Set communication parameters on opened port."""
        if self._socket is None:
            raise SerialException("Can only operate on open ports")

        # if self._timeout != 0 and self._interCharTimeout is not None:
            # XXX

        if self._write_timeout is not None:
            raise NotImplementedError('write_timeout is currently not supported')
            # XXX

        # Setup the connection
        # to get good performance, all parameter changes are sent first...
        if not 0 < self._baudrate < 2 ** 32:
            raise ValueError("invalid baudrate: {!r}".format(self._baudrate))
        self._rfc2217_port_settings['baudrate'].set(struct.pack(b'!I', self._baudrate))
        self._rfc2217_port_settings['datasize'].set(struct.pack(b'!B', self._bytesize))
        self._rfc2217_port_settings['parity'].set(struct.pack(b'!B', RFC2217_PARITY_MAP[self._parity]))
        self._rfc2217_port_settings['stopsize'].set(struct.pack(b'!B', RFC2217_STOPBIT_MAP[self._stopbits]))

        # and now wait until parameters are active
        items = self._rfc2217_port_settings.values()
        if self.logger:
            self.logger.debug("Negotiating settings: {}".format(items))
        timeout = Timeout(self._network_timeout)
        while not timeout.expired():
            time.sleep(0.05)    # prevent 100% CPU load
            if sum(o.active for o in items) == len(items):
                break
        else:
            raise SerialException("Remote does not accept parameter change (RFC2217): {!r}".format(items))
        if self.logger:
            self.logger.info("Negotiated settings: {}".format(items))

        if self._rtscts and self._xonxoff:
            raise ValueError('xonxoff and rtscts together are not supported')
        elif self._rtscts:
            self.rfc2217_set_control(SET_CONTROL_USE_HW_FLOW_CONTROL)
        elif self._xonxoff:
            self.rfc2217_set_control(SET_CONTROL_USE_SW_FLOW_CONTROL)
        else:
            self.rfc2217_set_control(SET_CONTROL_USE_NO_FLOW_CONTROL)

    def close(self):
        """Close port"""
        self.is_open = False
        if self._socket:
            try:
                self._socket.shutdown(socket.SHUT_RDWR)
                self._socket.close()
            except:
                # ignore errors.
                pass
        if self._thread:
            self._thread.join(7)  # XXX more than socket timeout
            self._thread = None
            # in case of quick reconnects, give the server some time
            time.sleep(0.3)
        self._socket = None

    def from_url(self, url):
        """\
        extract host and port from an URL string, other settings are extracted
        an stored in instance
        """
        parts = urlparse.urlsplit(url)
        if parts.scheme != "rfc2217":
            raise SerialException(
                'expected a string in the form '
                '"rfc2217://<host>:<port>[?option[&option...]]": '
                'not starting with rfc2217:// ({!r})'.format(parts.scheme))
        try:
            # process options now, directly altering self
            for option, values in urlparse.parse_qs(parts.query, True).items():
                if option == 'logging':
                    logging.basicConfig()   # XXX is that good to call it here?
                    self.logger = logging.getLogger('pySerial.rfc2217')
                    self.logger.setLevel(LOGGER_LEVELS[values[0]])
                    self.logger.debug('enabled logging')
                elif option == 'ign_set_control':
                    self._ignore_set_control_answer = True
                elif option == 'poll_modem':
                    self._poll_modem_state = True
                elif option == 'timeout':
                    self._network_timeout = float(values[0])
                else:
                    raise ValueError('unknown option: {!r}'.format(option))
            if not 0 <= parts.port < 65536:
                raise ValueError("port not in range 0...65535")
        except ValueError as e:
            raise SerialException(
                'expected a string in the form '
                '"rfc2217://<host>:<port>[?option[&option...]]": {}'.format(e))
        return (parts.hostname, parts.port)

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    @property
    def in_waiting(self):
        """Return the number of bytes currently in the input buffer."""
        if not self.is_open:
            raise PortNotOpenError()
        return self._read_buffer.qsize()

    def read(self, size=1):
        """\
        Read size bytes from the serial port. If a timeout is set it may
        return less characters as requested. With no timeout it will block
        until the requested number of bytes is read.
        """
        if not self.is_open:
            raise PortNotOpenError()
        data = bytearray()
        try:
            timeout = Timeout(self._timeout)
            while len(data) < size:
                if self._thread is None or not self._thread.is_alive():
                    raise SerialException('connection failed (reader thread died)')
                buf = self._read_buffer.get(True, timeout.time_left())
                if buf is None:
                    return bytes(data)
                data += buf
                if timeout.expired():
                    break
        except Queue.Empty:  # -> timeout
            pass
        return bytes(data)

    def write(self, data):
        """\
        Output the given byte string over the serial port. Can block if the
        connection is blocked. May raise SerialException if the connection is
        closed.
        """
        if not self.is_open:
            raise PortNotOpenError()
        with self._write_lock:
            try:
                self._socket.sendall(to_bytes(data).replace(IAC, IAC_DOUBLED))
            except socket.error as e:
                raise SerialException("connection failed (socket error): {}".format(e))
        return len(data)

    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""
        if not self.is_open:
            raise PortNotOpenError()
        self.rfc2217_send_purge(PURGE_RECEIVE_BUFFER)
        # empty read buffer
        while self._read_buffer.qsize():
            self._read_buffer.get(False)

    def reset_output_buffer(self):
        """\
        Clear output buffer, aborting the current output and
        discarding all that is in the buffer.
        """
        if not self.is_open:
            raise PortNotOpenError()
        self.rfc2217_send_purge(PURGE_TRANSMIT_BUFFER)

    def _update_break_state(self):
        """\
        Set break: Controls TXD. When active, to transmitting is
        possible.
        """
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('set BREAK to {}'.format('active' if self._break_state else 'inactive'))
        if self._break_state:
            self.rfc2217_set_control(SET_CONTROL_BREAK_ON)
        else:
            self.rfc2217_set_control(SET_CONTROL_BREAK_OFF)

    def _update_rts_state(self):
        """Set terminal status line: Request To Send."""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('set RTS to {}'.format('active' if self._rts_state else 'inactive'))
        if self._rts_state:
            self.rfc2217_set_control(SET_CONTROL_RTS_ON)
        else:
            self.rfc2217_set_control(SET_CONTROL_RTS_OFF)

    def _update_dtr_state(self):
        """Set terminal status line: Data Terminal Ready."""
        if not self.is_open:
            raise PortNotOpenError()
        if self.logger:
            self.logger.info('set DTR to {}'.format('active' if self._dtr_state else 'inactive'))
        if self._dtr_state:
            self.rfc2217_set_control(SET_CONTROL_DTR_ON)
        else:
            self.rfc2217_set_control(SET_CONTROL_DTR_OFF)

    @property
    def cts(self):
        """Read terminal status line: Clear To Send."""
        if not self.is_open:
            raise PortNotOpenError()
        return bool(self.get_modem_state() & MODEMSTATE_MASK_CTS)

    @property
    def dsr(self):
        """Read terminal status line: Data Set Ready."""
        if not self.is_open:
            raise PortNotOpenError()
        return bool(self.get_modem_state() & MODEMSTATE_MASK_DSR)

    @property
    def ri(self):
        """Read terminal status line: Ring Indicator."""
        if not self.is_open:
            raise PortNotOpenError()
        return bool(self.get_modem_state() & MODEMSTATE_MASK_RI)

    @property
    def cd(self):
        """Read terminal status line: Carrier Detect."""
        if not self.is_open:
            raise PortNotOpenError()
        return bool(self.get_modem_state() & MODEMSTATE_MASK_CD)

    # - - - platform specific - - -
    # None so far

    # - - - RFC2217 specific - - -

    def _telnet_read_loop(self):
        """Read loop for the socket."""
        mode = M_NORMAL
        suboption = None
        try:
            while self.is_open:
                try:
                    data = self._socket.recv(1024)
                except socket.timeout:
                    # just need to get out of recv form time to time to check if
                    # still alive
                    continue
                except socket.error as e:
                    # connection fails -> terminate loop
                    if self.logger:
                        self.logger.debug("socket error in reader thread: {}".format(e))
                    self._read_buffer.put(None)
                    break
                if not data:
                    self._read_buffer.put(None)
                    break  # lost connection
                for byte in iterbytes(data):
                    if mode == M_NORMAL:
                        # interpret as command or as data
                        if byte == IAC:
                            mode = M_IAC_SEEN
                        else:
                            # store data in read buffer or sub option buffer
                            # depending on state
                            if suboption is not None:
                                suboption += byte
                            else:
                                self._read_buffer.put(byte)
                    elif mode == M_IAC_SEEN:
                        if byte == IAC:
                            # interpret as command doubled -> insert character
                            # itself
                            if suboption is not None:
                                suboption += IAC
                            else:
                                self._read_buffer.put(IAC)
                            mode = M_NORMAL
                        elif byte == SB:
                            # sub option start
                            suboption = bytearray()
                            mode = M_NORMAL
                        elif byte == SE:
                            # sub option end -> process it now
                            self._telnet_process_subnegotiation(bytes(suboption))
                            suboption = None
                            mode = M_NORMAL
                        elif byte in (DO, DONT, WILL, WONT):
                            # negotiation
                            telnet_command = byte
                            mode = M_NEGOTIATE
                        else:
                            # other telnet commands
                            self._telnet_process_command(byte)
                            mode = M_NORMAL
                    elif mode == M_NEGOTIATE:  # DO, DONT, WILL, WONT was received, option now following
                        self._telnet_negotiate_option(telnet_command, byte)
                        mode = M_NORMAL
        finally:
            if self.logger:
                self.logger.debug("read thread terminated")

    # - incoming telnet commands and options

    def _telnet_process_command(self, command):
        """Process commands other than DO, DONT, WILL, WONT."""
        # Currently none. RFC2217 only uses negotiation and subnegotiation.
        if self.logger:
            self.logger.warning("ignoring Telnet command: {!r}".format(command))

    def _telnet_negotiate_option(self, command, option):
        """Process incoming DO, DONT, WILL, WONT."""
        # check our registered telnet options and forward command to them
        # they know themselves if they have to answer or not
        known = False
        for item in self._telnet_options:
            # can have more than one match! as some options are duplicated for
            # 'us' and 'them'
            if item.option == option:
                item.process_incoming(command)
                known = True
        if not known:
            # handle unknown options
            # only answer to positive requests and deny them
            if command == WILL or command == DO:
                self.telnet_send_option((DONT if command == WILL else WONT), option)
                if self.logger:
                    self.logger.warning("rejected Telnet option: {!r}".format(option))

    def _telnet_process_subnegotiation(self, suboption):
        """Process subnegotiation, the data between IAC SB and IAC SE."""
        if suboption[0:1] == COM_PORT_OPTION:
            if suboption[1:2] == SERVER_NOTIFY_LINESTATE and len(suboption) >= 3:
                self._linestate = ord(suboption[2:3])  # ensure it is a number
                if self.logger:
                    self.logger.info("NOTIFY_LINESTATE: {}".format(self._linestate))
            elif suboption[1:2] == SERVER_NOTIFY_MODEMSTATE and len(suboption) >= 3:
                self._modemstate = ord(suboption[2:3])  # ensure it is a number
                if self.logger:
                    self.logger.info("NOTIFY_MODEMSTATE: {}".format(self._modemstate))
                # update time when we think that a poll would make sense
                self._modemstate_timeout.restart(0.3)
            elif suboption[1:2] == FLOWCONTROL_SUSPEND:
                self._remote_suspend_flow = True
            elif suboption[1:2] == FLOWCONTROL_RESUME:
                self._remote_suspend_flow = False
            else:
                for item in self._rfc2217_options.values():
                    if item.ack_option == suboption[1:2]:
                        #~ print "processing COM_PORT_OPTION: %r" % list(suboption[1:])
                        item.check_answer(bytes(suboption[2:]))
                        break
                else:
                    if self.logger:
                        self.logger.warning("ignoring COM_PORT_OPTION: {!r}".format(suboption))
        else:
            if self.logger:
                self.logger.warning("ignoring subnegotiation: {!r}".format(suboption))

    # - outgoing telnet commands and options

    def _internal_raw_write(self, data):
        """internal socket write with no data escaping. used to send telnet stuff."""
        with self._write_lock:
            self._socket.sendall(data)

    def telnet_send_option(self, action, option):
        """Send DO, DONT, WILL, WONT."""
        self._internal_raw_write(IAC + action + option)

    def rfc2217_send_subnegotiation(self, option, value=b''):
        """Subnegotiation of RFC2217 parameters."""
        value = value.replace(IAC, IAC_DOUBLED)
        self._internal_raw_write(IAC + SB + COM_PORT_OPTION + option + value + IAC + SE)

    def rfc2217_send_purge(self, value):
        """\
        Send purge request to the remote.
        (PURGE_RECEIVE_BUFFER / PURGE_TRANSMIT_BUFFER / PURGE_BOTH_BUFFERS)
        """
        item = self._rfc2217_options['purge']
        item.set(value)  # transmit desired purge type
        item.wait(self._network_timeout)  # wait for acknowledge from the server

    def rfc2217_set_control(self, value):
        """transmit change of control line to remote"""
        item = self._rfc2217_options['control']
        item.set(value)  # transmit desired control type
        if self._ignore_set_control_answer:
            # answers are ignored when option is set. compatibility mode for
            # servers that answer, but not the expected one... (or no answer
            # at all) i.e. sredird
            time.sleep(0.1)  # this helps getting the unit tests passed
        else:
            item.wait(self._network_timeout)  # wait for acknowledge from the server

    def rfc2217_flow_server_ready(self):
        """\
        check if server is ready to receive data. block for some time when
        not.
        """
        #~ if self._remote_suspend_flow:
        #~     wait---

    def get_modem_state(self):
        """\
        get last modem state (cached value. If value is "old", request a new
        one. This cache helps that we don't issue to many requests when e.g. all
        status lines, one after the other is queried by the user (CTS, DSR
        etc.)
        """
        # active modem state polling enabled? is the value fresh enough?
        if self._poll_modem_state and self._modemstate_timeout.expired():
            if self.logger:
                self.logger.debug('polling modem state')
            # when it is older, request an update
            self.rfc2217_send_subnegotiation(NOTIFY_MODEMSTATE)
            timeout = Timeout(self._network_timeout)
            while not timeout.expired():
                time.sleep(0.05)    # prevent 100% CPU load
                # when expiration time is updated, it means that there is a new
                # value
                if not self._modemstate_timeout.expired():
                    break
            else:
                if self.logger:
                    self.logger.warning('poll for modem state failed')
            # even when there is a timeout, do not generate an error just
            # return the last known value. this way we can support buggy
            # servers that do not respond to polls, but send automatic
            # updates.
        if self._modemstate is not None:
            if self.logger:
                self.logger.debug('using cached modem state')
            return self._modemstate
        else:
            # never received a notification from the server
            raise SerialException("remote sends no NOTIFY_MODEMSTATE")


#############################################################################
# The following is code that helps implementing an RFC 2217 server.

class PortManager(object):
    """\
    This class manages the state of Telnet and RFC 2217. It needs a serial
    instance and a connection to work with. Connection is expected to implement
    a (thread safe) write function, that writes the string to the network.
    """

    def __init__(self, serial_port, connection, logger=None):
        self.serial = serial_port
        self.connection = connection
        self.logger = logger
        self._client_is_rfc2217 = False

        # filter state machine
        self.mode = M_NORMAL
        self.suboption = None
        self.telnet_command = None

        # states for modem/line control events
        self.modemstate_mask = 255
        self.last_modemstate = None
        self.linstate_mask = 0

        # all supported telnet options
        self._telnet_options = [
            TelnetOption(self, 'ECHO', ECHO, WILL, WONT, DO, DONT, REQUESTED),
            TelnetOption(self, 'we-SGA', SGA, WILL, WONT, DO, DONT, REQUESTED),
            TelnetOption(self, 'they-SGA', SGA, DO, DONT, WILL, WONT, INACTIVE),
            TelnetOption(self, 'we-BINARY', BINARY, WILL, WONT, DO, DONT, INACTIVE),
            TelnetOption(self, 'they-BINARY', BINARY, DO, DONT, WILL, WONT, REQUESTED),
            TelnetOption(self, 'we-RFC2217', COM_PORT_OPTION, WILL, WONT, DO, DONT, REQUESTED, self._client_ok),
            TelnetOption(self, 'they-RFC2217', COM_PORT_OPTION, DO, DONT, WILL, WONT, INACTIVE, self._client_ok),
        ]

        # negotiate Telnet/RFC2217 -> send initial requests
        if self.logger:
            self.logger.debug("requesting initial Telnet/RFC 2217 options")
        for option in self._telnet_options:
            if option.state is REQUESTED:
                self.telnet_send_option(option.send_yes, option.option)
        # issue 1st modem state notification

    def _client_ok(self):
        """\
        callback of telnet option. It gets called when option is activated.
        This one here is used to detect when the client agrees on RFC 2217. A
        flag is set so that other functions like check_modem_lines know if the
        client is OK.
        """
        # The callback is used for we and they so if one party agrees, we're
        # already happy. it seems not all servers do the negotiation correctly
        # and i guess there are incorrect clients too.. so be happy if client
        # answers one or the other positively.
        self._client_is_rfc2217 = True
        if self.logger:
            self.logger.info("client accepts RFC 2217")
        # this is to ensure that the client gets a notification, even if there
        # was no change
        self.check_modem_lines(force_notification=True)

    # - outgoing telnet commands and options

    def telnet_send_option(self, action, option):
        """Send DO, DONT, WILL, WONT."""
        self.connection.write(IAC + action + option)

    def rfc2217_send_subnegotiation(self, option, value=b''):
        """Subnegotiation of RFC 2217 parameters."""
        value = value.replace(IAC, IAC_DOUBLED)
        self.connection.write(IAC + SB + COM_PORT_OPTION + option + value + IAC + SE)

    # - check modem lines, needs to be called periodically from user to
    # establish polling

    def check_modem_lines(self, force_notification=False):
        """\
        read control lines from serial port and compare the last value sent to remote.
        send updates on changes.
        """
        modemstate = (
            (self.serial.cts and MODEMSTATE_MASK_CTS) |
            (self.serial.dsr and MODEMSTATE_MASK_DSR) |
            (self.serial.ri and MODEMSTATE_MASK_RI) |
            (self.serial.cd and MODEMSTATE_MASK_CD))
        # check what has changed
        deltas = modemstate ^ (self.last_modemstate or 0)  # when last is None -> 0
        if deltas & MODEMSTATE_MASK_CTS:
            modemstate |= MODEMSTATE_MASK_CTS_CHANGE
        if deltas & MODEMSTATE_MASK_DSR:
            modemstate |= MODEMSTATE_MASK_DSR_CHANGE
        if deltas & MODEMSTATE_MASK_RI:
            modemstate |= MODEMSTATE_MASK_RI_CHANGE
        if deltas & MODEMSTATE_MASK_CD:
            modemstate |= MODEMSTATE_MASK_CD_CHANGE
        # if new state is different and the mask allows this change, send
        # notification. suppress notifications when client is not rfc2217
        if modemstate != self.last_modemstate or force_notification:
            if (self._client_is_rfc2217 and (modemstate & self.modemstate_mask)) or force_notification:
                self.rfc2217_send_subnegotiation(
                    SERVER_NOTIFY_MODEMSTATE,
                    to_bytes([modemstate & self.modemstate_mask]))
                if self.logger:
                    self.logger.info("NOTIFY_MODEMSTATE: {}".format(modemstate))
            # save last state, but forget about deltas.
            # otherwise it would also notify about changing deltas which is
            # probably not very useful
            self.last_modemstate = modemstate & 0xf0

    # - outgoing data escaping

    def escape(self, data):
        """\
        This generator function is for the user. All outgoing data has to be
        properly escaped, so that no IAC character in the data stream messes up
        the Telnet state machine in the server.

        socket.sendall(escape(data))
        """
        for byte in iterbytes(data):
            if byte == IAC:
                yield IAC
                yield IAC
            else:
                yield byte

    # - incoming data filter

    def filter(self, data):
        """\
        Handle a bunch of incoming bytes. This is a generator. It will yield
        all characters not of interest for Telnet/RFC 2217.

        The idea is that the reader thread pushes data from the socket through
        this filter:

        for byte in filter(socket.recv(1024)):
            # do things like CR/LF conversion/whatever
            # and write data to the serial port
            serial.write(byte)

        (socket error handling code left as exercise for the reader)
        """
        for byte in iterbytes(data):
            if self.mode == M_NORMAL:
                # interpret as command or as data
                if byte == IAC:
                    self.mode = M_IAC_SEEN
                else:
                    # store data in sub option buffer or pass it to our
                    # consumer depending on state
                    if self.suboption is not None:
                        self.suboption += byte
                    else:
                        yield byte
            elif self.mode == M_IAC_SEEN:
                if byte == IAC:
                    # interpret as command doubled -> insert character
                    # itself
                    if self.suboption is not None:
                        self.suboption += byte
                    else:
                        yield byte
                    self.mode = M_NORMAL
                elif byte == SB:
                    # sub option start
                    self.suboption = bytearray()
                    self.mode = M_NORMAL
                elif byte == SE:
                    # sub option end -> process it now
                    self._telnet_process_subnegotiation(bytes(self.suboption))
                    self.suboption = None
                    self.mode = M_NORMAL
                elif byte in (DO, DONT, WILL, WONT):
                    # negotiation
                    self.telnet_command = byte
                    self.mode = M_NEGOTIATE
                else:
                    # other telnet commands
                    self._telnet_process_command(byte)
                    self.mode = M_NORMAL
            elif self.mode == M_NEGOTIATE:  # DO, DONT, WILL, WONT was received, option now following
                self._telnet_negotiate_option(self.telnet_command, byte)
                self.mode = M_NORMAL

    # - incoming telnet commands and options

    def _telnet_process_command(self, command):
        """Process commands other than DO, DONT, WILL, WONT."""
        # Currently none. RFC2217 only uses negotiation and subnegotiation.
        if self.logger:
            self.logger.warning("ignoring Telnet command: {!r}".format(command))

    def _telnet_negotiate_option(self, command, option):
        """Process incoming DO, DONT, WILL, WONT."""
        # check our registered telnet options and forward command to them
        # they know themselves if they have to answer or not
        known = False
        for item in self._telnet_options:
            # can have more than one match! as some options are duplicated for
            # 'us' and 'them'
            if item.option == option:
                item.process_incoming(command)
                known = True
        if not known:
            # handle unknown options
            # only answer to positive requests and deny them
            if command == WILL or command == DO:
                self.telnet_send_option((DONT if command == WILL else WONT), option)
                if self.logger:
                    self.logger.warning("rejected Telnet option: {!r}".format(option))

    def _telnet_process_subnegotiation(self, suboption):
        """Process subnegotiation, the data between IAC SB and IAC SE."""
        if suboption[0:1] == COM_PORT_OPTION:
            if self.logger:
                self.logger.debug('received COM_PORT_OPTION: {!r}'.format(suboption))
            if suboption[1:2] == SET_BAUDRATE:
                backup = self.serial.baudrate
                try:
                    (baudrate,) = struct.unpack(b"!I", suboption[2:6])
                    if baudrate != 0:
                        self.serial.baudrate = baudrate
                except ValueError as e:
                    if self.logger:
                        self.logger.error("failed to set baud rate: {}".format(e))
                    self.serial.baudrate = backup
                else:
                    if self.logger:
                        self.logger.info("{} baud rate: {}".format('set' if baudrate else 'get', self.serial.baudrate))
                self.rfc2217_send_subnegotiation(SERVER_SET_BAUDRATE, struct.pack(b"!I", self.serial.baudrate))
            elif suboption[1:2] == SET_DATASIZE:
                backup = self.serial.bytesize
                try:
                    (datasize,) = struct.unpack(b"!B", suboption[2:3])
                    if datasize != 0:
                        self.serial.bytesize = datasize
                except ValueError as e:
                    if self.logger:
                        self.logger.error("failed to set data size: {}".format(e))
                    self.serial.bytesize = backup
                else:
                    if self.logger:
                        self.logger.info("{} data size: {}".format('set' if datasize else 'get', self.serial.bytesize))
                self.rfc2217_send_subnegotiation(SERVER_SET_DATASIZE, struct.pack(b"!B", self.serial.bytesize))
            elif suboption[1:2] == SET_PARITY:
                backup = self.serial.parity
                try:
                    parity = struct.unpack(b"!B", suboption[2:3])[0]
                    if parity != 0:
                        self.serial.parity = RFC2217_REVERSE_PARITY_MAP[parity]
                except ValueError as e:
                    if self.logger:
                        self.logger.error("failed to set parity: {}".format(e))
                    self.serial.parity = backup
                else:
                    if self.logger:
                        self.logger.info("{} parity: {}".format('set' if parity else 'get', self.serial.parity))
                self.rfc2217_send_subnegotiation(
                    SERVER_SET_PARITY,
                    struct.pack(b"!B", RFC2217_PARITY_MAP[self.serial.parity]))
            elif suboption[1:2] == SET_STOPSIZE:
                backup = self.serial.stopbits
                try:
                    stopbits = struct.unpack(b"!B", suboption[2:3])[0]
                    if stopbits != 0:
                        self.serial.stopbits = RFC2217_REVERSE_STOPBIT_MAP[stopbits]
                except ValueError as e:
                    if self.logger:
                        self.logger.error("failed to set stop bits: {}".format(e))
                    self.serial.stopbits = backup
                else:
                    if self.logger:
                        self.logger.info("{} stop bits: {}".format('set' if stopbits else 'get', self.serial.stopbits))
                self.rfc2217_send_subnegotiation(
                    SERVER_SET_STOPSIZE,
                    struct.pack(b"!B", RFC2217_STOPBIT_MAP[self.serial.stopbits]))
            elif suboption[1:2] == SET_CONTROL:
                if suboption[2:3] == SET_CONTROL_REQ_FLOW_SETTING:
                    if self.serial.xonxoff:
                        self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_USE_SW_FLOW_CONTROL)
                    elif self.serial.rtscts:
                        self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_USE_HW_FLOW_CONTROL)
                    else:
                        self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_USE_NO_FLOW_CONTROL)
                elif suboption[2:3] == SET_CONTROL_USE_NO_FLOW_CONTROL:
                    self.serial.xonxoff = False
                    self.serial.rtscts = False
                    if self.logger:
                        self.logger.info("changed flow control to None")
                    self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_USE_NO_FLOW_CONTROL)
                elif suboption[2:3] == SET_CONTROL_USE_SW_FLOW_CONTROL:
                    self.serial.xonxoff = True
                    if self.logger:
                        self.logger.info("changed flow control to XON/XOFF")
                    self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_USE_SW_FLOW_CONTROL)
                elif suboption[2:3] == SET_CONTROL_USE_HW_FLOW_CONTROL:
                    self.serial.rtscts = True
                    if self.logger:
                        self.logger.info("changed flow control to RTS/CTS")
                    self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_USE_HW_FLOW_CONTROL)
                elif suboption[2:3] == SET_CONTROL_REQ_BREAK_STATE:
                    if self.logger:
                        self.logger.warning("requested break state - not implemented")
                    pass  # XXX needs cached value
                elif suboption[2:3] == SET_CONTROL_BREAK_ON:
                    self.serial.break_condition = True
                    if self.logger:
                        self.logger.info("changed BREAK to active")
                    self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_BREAK_ON)
                elif suboption[2:3] == SET_CONTROL_BREAK_OFF:
                    self.serial.break_condition = False
                    if self.logger:
                        self.logger.info("changed BREAK to inactive")
                    self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_BREAK_OFF)
                elif suboption[2:3] == SET_CONTROL_REQ_DTR:
                    if self.logger:
                        self.logger.warning("requested DTR state - not implemented")
                    pass  # XXX needs cached value
                elif suboption[2:3] == SET_CONTROL_DTR_ON:
                    self.serial.dtr = True
                    if self.logger:
                        self.logger.info("changed DTR to active")
                    self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_DTR_ON)
                elif suboption[2:3] == SET_CONTROL_DTR_OFF:
                    self.serial.dtr = False
                    if self.logger:
                        self.logger.info("changed DTR to inactive")
                    self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_DTR_OFF)
                elif suboption[2:3] == SET_CONTROL_REQ_RTS:
                    if self.logger:
                        self.logger.warning("requested RTS state - not implemented")
                    pass  # XXX needs cached value
                    #~ self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_RTS_ON)
                elif suboption[2:3] == SET_CONTROL_RTS_ON:
                    self.serial.rts = True
                    if self.logger:
                        self.logger.info("changed RTS to active")
                    self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_RTS_ON)
                elif suboption[2:3] == SET_CONTROL_RTS_OFF:
                    self.serial.rts = False
                    if self.logger:
                        self.logger.info("changed RTS to inactive")
                    self.rfc2217_send_subnegotiation(SERVER_SET_CONTROL, SET_CONTROL_RTS_OFF)
                #~ elif suboption[2:3] == SET_CONTROL_REQ_FLOW_SETTING_IN:
                #~ elif suboption[2:3] == SET_CONTROL_USE_NO_FLOW_CONTROL_IN:
                #~ elif suboption[2:3] == SET_CONTROL_USE_SW_FLOW_CONTOL_IN:
                #~ elif suboption[2:3] == SET_CONTROL_USE_HW_FLOW_CONTOL_IN:
                #~ elif suboption[2:3] == SET_CONTROL_USE_DCD_FLOW_CONTROL:
                #~ elif suboption[2:3] == SET_CONTROL_USE_DTR_FLOW_CONTROL:
                #~ elif suboption[2:3] == SET_CONTROL_USE_DSR_FLOW_CONTROL:
            elif suboption[1:2] == NOTIFY_LINESTATE:
                # client polls for current state
                self.rfc2217_send_subnegotiation(
                    SERVER_NOTIFY_LINESTATE,
                    to_bytes([0]))   # sorry, nothing like that implemented
            elif suboption[1:2] == NOTIFY_MODEMSTATE:
                if self.logger:
                    self.logger.info("request for modem state")
                # client polls for current state
                self.check_modem_lines(force_notification=True)
            elif suboption[1:2] == FLOWCONTROL_SUSPEND:
                if self.logger:
                    self.logger.info("suspend")
                self._remote_suspend_flow = True
            elif suboption[1:2] == FLOWCONTROL_RESUME:
                if self.logger:
                    self.logger.info("resume")
                self._remote_suspend_flow = False
            elif suboption[1:2] == SET_LINESTATE_MASK:
                self.linstate_mask = ord(suboption[2:3])  # ensure it is a number
                if self.logger:
                    self.logger.info("line state mask: 0x{:02x}".format(self.linstate_mask))
            elif suboption[1:2] == SET_MODEMSTATE_MASK:
                self.modemstate_mask = ord(suboption[2:3])  # ensure it is a number
                if self.logger:
                    self.logger.info("modem state mask: 0x{:02x}".format(self.modemstate_mask))
            elif suboption[1:2] == PURGE_DATA:
                if suboption[2:3] == PURGE_RECEIVE_BUFFER:
                    self.serial.reset_input_buffer()
                    if self.logger:
                        self.logger.info("purge in")
                    self.rfc2217_send_subnegotiation(SERVER_PURGE_DATA, PURGE_RECEIVE_BUFFER)
                elif suboption[2:3] == PURGE_TRANSMIT_BUFFER:
                    self.serial.reset_output_buffer()
                    if self.logger:
                        self.logger.info("purge out")
                    self.rfc2217_send_subnegotiation(SERVER_PURGE_DATA, PURGE_TRANSMIT_BUFFER)
                elif suboption[2:3] == PURGE_BOTH_BUFFERS:
                    self.serial.reset_input_buffer()
                    self.serial.reset_output_buffer()
                    if self.logger:
                        self.logger.info("purge both")
                    self.rfc2217_send_subnegotiation(SERVER_PURGE_DATA, PURGE_BOTH_BUFFERS)
                else:
                    if self.logger:
                        self.logger.error("undefined PURGE_DATA: {!r}".format(list(suboption[2:])))
            else:
                if self.logger:
                    self.logger.error("undefined COM_PORT_OPTION: {!r}".format(list(suboption[1:])))
        else:
            if self.logger:
                self.logger.warning("unknown subnegotiation: {!r}".format(suboption))


# simple client test
if __name__ == '__main__':
    import sys
    s = Serial('rfc2217://localhost:7000', 115200)
    sys.stdout.write('{}\n'.format(s))

    sys.stdout.write("write...\n")
    s.write(b"hello\n")
    s.flush()
    sys.stdout.write("read: {}\n".format(s.read(5)))
    s.close()



import cfgtools

from util import find_key_value


class ZoomLevel(object):
    def __init__(self, group):
        # type: (cfgtools.Group) -> None

        self.name = ''
        self.cols = 0
        self.lines = 0
        self.rows = 0
        self.fontsize = 0

        self.parse_zoom_level(group)

    def parse_zoom_level(self, zoom_level_group):
        # type: (cfgtools.Group) -> ()
        self.name = zoom_level_group.name
        self.cols = find_key_value(zoom_level_group, "Cols", 0)
        self.lines = find_key_value(zoom_level_group, "Lines", 0)
        self.rows = find_key_value(zoom_level_group, "Rows", 0)
        self.fontsize = find_key_value(zoom_level_group, "FontSize", 0)



import sys
import os
import iso8601
from typing import List
from os import listdir
from os.path import isfile, join


class LogLine(object):
    def __init__(self, thread_number, order_id, line_content):
        self.thread_number = thread_number
        self.order_id = order_id
        self.line_content = line_content


def read_log_files(log_files):
    file_data = ""
    for file_name in log_files:
        with open(file_name, "rb") as f:
            file_data += f.read()

    return file_data.decode("utf-8")


def get_thread_number(line):
    return line.split("-")[4].strip()


def get_order_id(line_content):
    #  type: (str) -> int
    index = line_content.index("\n")
    if index < 0:
        return -1
    index2 = line_content.index("\n", index + 1)
    order_header_line = line_content[index + 1:index2]
    comps = order_header_line.split(",")
    data = comps[0].split(":")
    try:
        return int(data[1].strip())
    except:
        return -1


def get_log_lines(log_data):
    lines = []
    line_content = ""
    first_line = None
    for line in log_data.split('\n'):
        line_start = line[0:23]
        try:
            iso8601.parse_date(line_start)
            if line_content != "":
                lines.append(LogLine(get_thread_number(first_line),
                                     get_order_id(line_content),
                                     line_content))

            first_line = line
            line_content = ""
        except:
            pass

        line_content += line + "\n"

    if line_content != "":
        lines.append(LogLine(get_thread_number(first_line), get_order_id(line_content), line_content))

    return lines


def separate_lines_by_thread(log_lines):
    # type: (List[LogLine]) -> List[List[LogLine]]
    line_dict = {}
    for line in log_lines:
        if line.thread_number not in line_dict:
            line_dict[line.thread_number] = []
        line_dict[line.thread_number].append(line)

    return line_dict.values()


def separate_lines_by_order(log_lines):
    # type: (List[LogLine]) -> List[List[LogLine]]
    line_dict = {}
    for line in log_lines:
        if line.order_id not in line_dict:
            line_dict[line.order_id] = []
        line_dict[line.order_id].append(line)

    return line_dict


def main():
    log_directory = os.path.abspath(sys.argv[1])
    log_files = [os.path.join(log_directory, f) for f in listdir(log_directory) if isfile(join(log_directory, f)) and ".log" in f]
    log_lines = get_log_lines(read_log_files(log_files))

    if len(sys.argv) > 2:
        order_id = int(sys.argv[2])
        lines_by_order = separate_lines_by_order(log_lines)
        if order_id not in lines_by_order:
            print("Order {} not found in files".format(order_id))
            return -1

        order_lines = lines_by_order[order_id]
        with open(os.path.join(log_directory, "{}.log".format(order_id)), "wb") as log_file:
            for line in order_lines:
                log_file.write(line.line_content.encode("utf-8"))
    else:
        separated_lines = separate_lines_by_thread(log_lines)

        index = 0
        for thread_logs in separated_lines:
            with open(os.path.join(log_directory, "log{}.log".format(index)), "wb") as log_file:
                for line in thread_logs:
                    log_file.write(line.line_content.encode("utf-8"))
            index += 1


main()



# nace.py - functions for handling EU NACE classification
# coding: utf-8
#
# Copyright (C) 2017-2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NACE (classification for businesses in the European Union).

The NACE (nomenclature statistique des activités économiques dans la
Communauté européenne) is a 4-level (and up to 4 digit) code for classifying
economic activities. It is the European implementation of the UN
classification ISIC.

The first 4 digits are the same in all EU countries while additional levels
and digits may be vary between countries. This module validates the numbers
according to revision 2 and based on the registry as published by the EC.

More information:

* https://en.wikipedia.org/wiki/Statistical_Classification_of_Economic_Activities_in_the_European_Community
* https://ec.europa.eu/eurostat/ramon/nomenclatures/index.cfm?TargetUrl=LST_NOM_DTL&StrNom=NACE_REV2&StrLanguageCode=EN&IntPcKey=&StrLayoutCode=HIERARCHIC

>>> validate('A')
'A'
>>> validate('62.01')
'6201'
>>> str(get_label('62.01'))
'Computer programming activities'
>>> validate('62.05')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('62059')  # does not validate country-specific numbers
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('6201')
'62.01'
"""  # noqa: E501

import warnings

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, '.').strip()


def info(number):
    """Lookup information about the specified NACE. This returns a dict."""
    number = compact(number)
    from stdnum import numdb
    info = dict()
    for _n, i in numdb.get('eu/nace').info(number):
        if not i:
            raise InvalidComponent()
        info.update(i)
    return info


def get_label(number):
    """Lookup the category label for the number."""
    return info(number)['label']


def label(number):  # pragma: no cover (deprecated function)
    """DEPRECATED: use `get_label()` instead."""  # noqa: D40
    warnings.warn(
        'label() has been to get_label()',
        DeprecationWarning, stacklevel=2)
    return get_label(number)


def validate(number):
    """Check if the number is a valid NACE. This checks the format and
    searches the registry to see if it exists."""
    number = compact(number)
    if len(number) > 4:
        raise InvalidLength()
    elif len(number) == 1:
        if not number.isalpha():
            raise InvalidFormat()
    else:
        if not isdigits(number):
            raise InvalidFormat()
    info(number)
    return number


def is_valid(number):
    """Check if the number is a valid NACE. This checks the format and
    searches the registry to see if it exists."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return '.'.join((number[:2], number[2:])).strip('.')



import os

from systools import sys_parsefilepath

from application.model import UpdateType
from application.model.customexception import ErrorUpdatingApache
from application.service import UpdaterService
from application.util import (
    LoggerUtil,
    settings,
)

logger = LoggerUtil.get_logger_name(__name__)


class ApacheUpdaterService(object):
    def __init__(
        self,
        updater_service,  # type: UpdaterService
    ):
        # type: (...) -> None

        hv_data_dir = sys_parsefilepath("{HVDATADIR}")

        self.update_type = UpdateType.binaries
        self.update_name = self.update_type.name
        self.updater_service = updater_service
        self.apache_path = os.path.join(
            hv_data_dir,
            "..",
            "..",
            "apache",
        )
        self.apache_conf_path = os.path.join(
            self.apache_path,
            "conf",
        )
        self.apache_conf = os.path.join(
            self.apache_conf_path,
            "httpd.conf",
        )
        self.apache_conf_bkp = os.path.join(
            self.apache_conf_path,
            "httpd.conf.bkp",
        )
        self.apache_conf_bkp = os.path.join(
            self.apache_conf_path,
            "httpd.conf.bkp",
        )
        self.apache_logs_path = os.path.join(
            self.apache_path,
            "logs",
        )
        self.apache_access_log = os.path.join(
            self.apache_logs_path,
            "access.log",
        )
        self.lines_to_uncomment = [
            "#LoadModule deflate_module modules/mod_deflate.so",
            "#LoadModule filter_module modules/mod_filter.so",
            "#LoadModule headers_module modules/mod_headers.so",
        ]
        self.lines_to_comment = [
            'CustomLog "logs/access.log" common',
        ]

        self.deflate_module = "\n<IfModule deflate_module>\n"
        self.deflate_module += "DeflateCompressionLevel 1\n"
        self.deflate_module += "AddOutputFilterByType DEFLATE application/javascript application/rss+xml "
        self.deflate_module += "application/vnd.ms-fontobject application/x-font application/x-font-opentype "
        self.deflate_module += "application/x-font-otf application/x-font-truetype application/x-font-ttf "
        self.deflate_module += "application/x-javascript application/xhtml+xml application/xml font/opentype "
        self.deflate_module += "font/otf font/ttf image/svg+xml image/x-icon text/css text/html text/javascript "
        self.deflate_module += "text/plain text/xml\n"
        self.deflate_module += "</IfModule>\n"

        self.files_config = "<Files index.html>\n"
        self.files_config += "FileETag None\n"
        self.files_config += "Header unset ETag\n"
        self.files_config += """Header set Cache-Control "max-age=0, no-cache, no-store, must-revalidate"\n"""
        self.files_config += """Header set Pragma "no-cache"\n"""
        self.files_config += "Header set Expires 0\n"
        self.files_config += "</Files>\n"

        self.files_match = """\n<FilesMatch "\\.(jpg|jpeg|png|gif|json|ico|mp3|mp4|ttf)$">\n"""
        self.files_match += """Header set Cache-Control "max-age=0, must-revalidate"\n"""
        self.files_match += "</FilesMatch>\n"

    def update_httpd_conf(self):
        # type: (...) -> None

        settings.set_update_control(
            self.update_name,
            "$UPDATING_APACHE_HTTPD_CONF",
        )
        try:
            logger.info("Updating httpd.conf file.")
            if os.path.exists(self.apache_conf):
                self._create_backup()
                data = self._read_apache_conf()
                configs_to_remove = []
                index_deflate_module = -1
                for i, line in enumerate(data):
                    original_line = line
                    line = line.strip()
                    if line.startswith("#LoadModule"):
                        if line in self.lines_to_uncomment:
                            data[i] = line[1:] + "\n"
                        else:
                            continue
                    elif line.startswith("LoadModule") or line.startswith("#") or line == "":
                        continue
                    elif line in self.lines_to_comment:
                        data[i] = self._get_line_indentation(line=original_line) + "#" + line + "\n"
                    elif line == "LogLevel warn":
                        index_deflate_module = i + 1
                    elif line in ["<IfModule deflate_module>", "<Files index.html>"]:
                        configs_to_remove.append(i)
                    elif line.startswith("<FilesMatch"):
                        configs_to_remove.append(i)

                data = self._remove_configs(
                    data=data,
                    indexes=configs_to_remove,
                )
                data.insert(
                    index_deflate_module,
                    self.deflate_module,
                )
                data.append(self.files_config)
                data.append(self.files_match)
                data = list(
                    filter(
                        None,
                        data,
                    )
                )
                self._write_apache_conf(data=data)
                logger.info("Updated httpd.conf successfully!")
        except Exception as _:
            logger.exception("Error updating httpd.conf file.")
            self._restore_apache_conf()
            raise ErrorUpdatingApache()
        finally:
            self.updater_service.delete_file(path=self.apache_conf_bkp)
            self._clean_access_log_file()

    def _clean_access_log_file(
        self,
    ):
        # type: (...) -> None

        try:
            self.updater_service.delete_file(self.apache_access_log)
        except Exception as _:
            logger.exception("Error cleaning access.log file.")

    def _restore_apache_conf(self):
        try:
            logger.info("Restoring httpd.conf backup.")
            self.updater_service.delete_file(self.apache_conf)
            self.updater_service.copy_file(
                original_file_path=self.apache_conf_bkp,
                destination_file_path=self.apache_conf,
            )
            logger.info("Restored httpd.conf successfully!")
        except Exception as _:
            logger.exception("Error restoring httpd.conf file.")

    def _create_backup(self):
        logger.info("Creating httpd.conf backup.")
        self.updater_service.copy_file(
            original_file_path=self.apache_conf,
            destination_file_path=self.apache_conf_bkp,
        )

    def _read_apache_conf(self):
        # type: (...) -> list

        with open(self.apache_conf, "r") as f:
            data = f.readlines()
        return data

    def _write_apache_conf(
        self,
        data,  # type: list
    ):
        # type: (...) -> None

        with open(self.apache_conf, "w") as f:
            f.writelines(data)

    @staticmethod
    def _remove_configs(
        data,     # type: list
        indexes,  # type: list
    ):
        # type: (...) -> list

        for i in indexes:
            tag = data[i].strip()[1:].split(" ")[0]
            config_end = "</" + tag + ">"
            j = i
            while True:
                line = data[j].strip()
                data[j] = ""
                if line == config_end:
                    if j + 1 < len(data) - 1 and data[j + 1] == "\n":
                        data[j + 1] = ""
                    break
                j += 1

        return data

    @staticmethod
    def _get_line_indentation(
        line,  # type: str
    ):
        # type: (...) -> str

        first_char = line[0]
        valid_chars = [" ", "\t"]
        indentation = ""
        if first_char in valid_chars:
            for c in line:
                if c == first_char:
                    indentation += c
                else:
                    break

        return indentation



from ._OperatorSession import OperatorSession
from ._Authenticator import Authenticator
from ._PosController import PosController
from ._OrderManager import OrderManager
from ._TableService import TableService
from ._Table import Table
from ._SaleLine import SaleLine
from ._TenderIdEnum import TenderIdEnum
from ._Order import Order
from ._CustomOrderProperty import CustomOrderProperty
from ._TablePictureToOrdersParser import TablePictureToOrdersParser
from ._PrintReport import PrintReport
from ._I18nImpl import I18nImpl
from ._I18n import I18n

__all__ = [
    "OperatorSession",
    "Authenticator",
    "PosController",
    "OrderManager",
    "TableService",
    "Table",
    "SaleLine",
    "TenderIdEnum",
    "Order",
    "CustomOrderProperty",
    "TablePictureToOrdersParser",
    "PrintReport",
    "I18nImpl",
    "I18n",
]



from messagebus import AckMessage, NakMessage
from messageprocessorutil import UuidMessageProcessor
from production.interactor import GetOrderMaxPrepTimeInteractor
from production.manager import ProductionManager
from production.model.request import GetOrderXmlRequest
from production.repository import ProductRepository


class GetOrderMaxPrepTimeProcessor(UuidMessageProcessor):
    def __init__(self, production_manager, product_repository):
        # type: (ProductionManager, ProductRepository) -> None
        super(GetOrderMaxPrepTimeProcessor, self).__init__()
        self.production_manager = production_manager
        self.product_repository = product_repository
        self.get_order_max_prep_time_interactor = GetOrderMaxPrepTimeInteractor(self.product_repository)

    def parse_data(self, data, event_type):
        return GetOrderXmlRequest(int(data))

    def call_business(self, request):
        production_order = self.production_manager.get_production_order(order_id=str(request.order_id))
        return self.get_order_max_prep_time_interactor.execute(production_order)

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(message, AckMessage(str(result).encode("utf-8")))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        exception_message = "Error obtaining order max prep time: {}".format(exception.message)
        message_bus.reply_message(message, NakMessage(exception_message))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        exception_message = "Error parsing received data: {}; Received data: {}".format(exception.message, data)
        message_bus.reply_message(message, NakMessage(exception_message))



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.exceptions import ClientError
from botocore.utils import get_service_module_name


class BaseClientExceptions(object):
    ClientError = ClientError

    def __init__(self, code_to_exception):
        """Base class for exceptions object on a client

        :type code_to_exception: dict
        :param code_to_exception: Mapping of error codes (strings) to exception
            class that should be raised when encountering a particular
            error code.
        """
        self._code_to_exception = code_to_exception

    def from_code(self, error_code):
        """Retrieves the error class based on the error code

        This is helpful for identifying the exception class needing to be
        caught based on the ClientError.parsed_reponse['Error']['Code'] value

        :type error_code: string
        :param error_code: The error code associated to a ClientError exception

        :rtype: ClientError or a subclass of ClientError
        :returns: The appropriate modeled exception class for that error
            code. If the error code does not match any of the known
            modeled exceptions then return a generic ClientError.
        """
        return self._code_to_exception.get(error_code, self.ClientError)

    def __getattr__(self, name):
        exception_cls_names = [
            exception_cls.__name__ for exception_cls
            in self._code_to_exception.values()
        ]
        raise AttributeError(
            '%r object has no attribute %r. Valid exceptions are: %s' % (
                self, name, ', '.join(exception_cls_names)))


class ClientExceptionsFactory(object):
    def __init__(self):
        self._client_exceptions_cache = {}

    def create_client_exceptions(self, service_model):
        """Creates a ClientExceptions object for the particular service client

        :type service_model: botocore.model.ServiceModel
        :param service_model: The service model for the client

        :rtype: object that subclasses from BaseClientExceptions
        :returns: The exceptions object of a client that can be used
            to grab the various different modeled exceptions.
        """
        service_name = service_model.service_name
        if service_name not in self._client_exceptions_cache:
            client_exceptions = self._create_client_exceptions(service_model)
            self._client_exceptions_cache[service_name] = client_exceptions
        return self._client_exceptions_cache[service_name]

    def _create_client_exceptions(self, service_model):
        cls_props = {}
        code_to_exception = {}
        for error_shape in service_model.error_shapes:
            exception_name = str(error_shape.name)
            exception_cls = type(exception_name, (ClientError,), {})
            cls_props[exception_name] = exception_cls
            code = str(error_shape.error_code)
            code_to_exception[code] = exception_cls
        cls_name = str(get_service_module_name(service_model) + 'Exceptions')
        client_exceptions_cls = type(
            cls_name, (BaseClientExceptions,), cls_props)
        return client_exceptions_cls(code_to_exception)



from requests import Request, Session, Response
from typing import Optional


class NfceRequest(object):
    def __init__(self, certificate_key_path, certificate_path, sefaz_certificate_path, check_situation_timeout):
        self.certificate_key_path = certificate_key_path
        self.certificate_path = certificate_path
        self.sefaz_certificate_path = sefaz_certificate_path
        self.check_situation_timeout = check_situation_timeout

    def envia_nfce(self, request, url, soap_action=None, timeout=None):
        # type: (str, str, Optional[str], Optional[int]) -> Response

        req = Request('POST', url, data=request)
        prepared_request = req.prepare()

        prepared_request.headers['Content-Type'] = 'application/soap+xml; charset=utf-8'
        if soap_action is not None:
            prepared_request.headers['SOAPAction'] = soap_action

        cert = (self.certificate_path, self.certificate_key_path)
        ca_cert = self.sefaz_certificate_path

        s = Session()
        resp = s.send(
            prepared_request,
            cert=cert,
            verify=ca_cert,
            timeout=timeout or self.check_situation_timeout
        )
        return resp



from sysactions import action

from .service_period_commons import get_selection_service_period, set_services_period_all_pos


@action
def do_service_period_pos(pos_id):
    # type: (int) -> bool

    service_period = get_selection_service_period(pos_id)
    if service_period is None:
        return False

    set_services_period_all_pos(pos_id, service_period)
    return True



import json

from messagebus import (
    AckMessage,
    DataType,
    NakMessage
)
from modelcommon.model import (
    SitefResponse,
    SitefResponseJsonEncoder
)
from typing import (
    Any,
    Union
)

from ._CommonProcessor import _CommonProcessor
from model import CancelPaymentOrderData


class CancelPaymentOrderProcessor(_CommonProcessor):
    def parse_data(
        self,
        data,       # type: str
        event_type  # type: str
    ):
        # type: (...) -> CancelPaymentOrderData

        data = json.loads(data or "{}")
        return CancelPaymentOrderData(data)

    def call_business(
        self,
        cancel_payment_order_data   # type: CancelPaymentOrderData
    ):
        # type: (...) -> Any[Union[str, SitefResponse]]

        processor = self.payment_router.get_payment_processor(cancel_payment_order_data.electronic_type)
        if processor:
            return processor.cancel_payment_order(cancel_payment_order_data=cancel_payment_order_data)

    def format_response(self, message_bus, message, event, input_obj, result):
        if isinstance(result, SitefResponse):
            result = json.dumps(result, cls=SitefResponseJsonEncoder)
            message_bus.reply_message(message, AckMessage(data=result, data_type=DataType.string))
        else:
            message_bus.reply_message(message, NakMessage(data=result, data_type=DataType.string))



# vat.py - functions for handling United Kingdom VAT numbers
#
# Copyright (C) 2012-2021 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""VAT (United Kingdom (and Isle of Man) VAT registration number).

The VAT number can either be a 9-digit standard number, a 12-digit standard
number followed by a 3-digit branch identifier, a 5-digit number for
government departments (first two digits are GD) or a 5-digit number for
health authorities (first two digits are HA). The 9-digit variants use a
weighted checksum.

>>> validate('GB 980 7806 84')
'980780684'
>>> validate('802311781')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('980780684')
'980 7806 84'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' -.').upper().strip()
    if number.startswith('GB') or number.startswith('XI'):
        number = number[2:]
    return number


def checksum(number):
    """Calculate the checksum. The checksum is only used for the 9 digits
    of the number and the result can either be 0 or 42."""
    weights = (8, 7, 6, 5, 4, 3, 2, 10, 1)
    return sum(w * int(n) for w, n in zip(weights, number)) % 97


def validate(number):
    """Check if the number is a valid VAT number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) == 5:
        if not isdigits(number[2:]):
            raise InvalidFormat()
        if number.startswith('GD') and int(number[2:]) < 500:
            # government department
            pass
        elif number.startswith('HA') and int(number[2:]) >= 500:
            # health authority
            pass
        else:
            raise InvalidComponent()
    elif len(number) == 11 and number[0:6] in ('GD8888', 'HA8888'):
        if not isdigits(number[6:]):
            raise InvalidFormat()
        if number.startswith('GD') and int(number[6:9]) < 500:
            # government department
            pass
        elif number.startswith('HA') and int(number[6:9]) >= 500:
            # health authority
            pass
        else:
            raise InvalidComponent()
        if int(number[6:9]) % 97 != int(number[9:11]):
            raise InvalidChecksum()
    elif len(number) in (9, 12):
        if not isdigits(number):
            raise InvalidFormat()
        # standard number: nnn nnnn nn
        # branch trader: nnn nnnn nn nnn (ignore the last thee digits)
        # restarting: 100 nnnn nn
        if int(number[:3]) >= 100:
            if checksum(number[:9]) not in (0, 42, 55):
                raise InvalidChecksum()
        else:
            if checksum(number[:9]) != 0:
                raise InvalidChecksum()
    else:
        raise InvalidLength()
    return number


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    if len(number) == 5:
        # government department or health authority
        return number
    if len(number) == 12:
        # includes branch number
        return number[:3] + ' ' + number[3:7] + ' ' + number[7:9] + ' ' + number[9:]
    # standard number: nnn nnnn nn
    return number[:3] + ' ' + number[3:7] + ' ' + number[7:]
